import pandas as pd

import dvcx.error
from dvcx.lib.dataset import Dataset
from dvcx.lib.webdataset import WebDataset
from dvcx.query.schema import C
from dvcx.sql import literal
from dvcx.sql.functions import array, greatest, least, string

name = "wds"
wds = Dataset(name=name)
try:
    df = wds.limit(3).to_pandas()
except dvcx.error.DatasetNotFoundError:
    (
        Dataset("gcs://dvcx-datacomp-small/shards", anon=True)
        .filter(C.name.glob("00000000.tar"))
        .generate(WebDataset())
        .save(name)
    )
    df = wds.limit(3).to_pandas()

print(df.columns.tolist())
columns = [
    "parent",
    "name",
    "vtype",
    "dir_type",
    "size",
    "caption",
    "url",
    "width",
    "height",
    "original_width",
    "original_height",
]
with pd.option_context("display.max_columns", None):
    print(df[columns])

filtered = (
    wds.filter(string.length(C.caption) > 5)
    .filter(array.length(string.split(C.caption, literal(" "))) > 2)
    .filter(least(C.original_width, C.original_height) > 200)
    .filter(
        greatest(C.original_width, C.original_height)
        / least(C.original_width, C.original_height)
        < 3.0
    )
)
filtered_df = filtered.limit(3).to_pandas()[columns]
with pd.option_context("display.max_columns", None):
    print(filtered_df)

print(f"wds count:      {wds.count():>6}")
print(f"filtered count: {filtered.count():>6}")



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from dvcx.lib.feature_types import FeatureLike


class FeatureReader(ABC):
    def __init__(self, fr_class: "FeatureLike"):
        """
        Class to call on feature values to perform post-processing. Used when
        iterating over dataset with `ds.to_pytorch()` and `ds.get_values()`.

        The class must include:
        - `self.fr_class` to define the feature class to read.
        - `self.__call__(self, value)` to call on the feature value returned by
          `self.fr_class.get_value()`.

        Examples:
            >>> class PrefixReader(FeatureReader):
            >>>     def __call__(self, value):
            >>>         return "prefix-" + value
            >>> for row in ds.get_values(PrefixReader(MyFeature)):
            >>>     print(row)

            >>> class SuffixReader(FeatureReader):
            >>>     def __init__(self, fr_class, suffix):
            >>>         self.suffix = suffix
            >>>         super().__init__(fr_class)
            >>>     def __call__(self, value):
            >>>         return value + self.suffix
            >>> for row in ds.get_values(SuffixReader(MyFeature, "-suffix")):
            >>>     print(row)
        """
        self.fr_class = fr_class

    @abstractmethod
    def __call__(self, value: Any) -> Any:
        pass


class LabelReader(FeatureReader):
    def __init__(self, fr_class: "FeatureLike", classes: list):
        """Get column values as 0-based integer index of classes."""
        self.classes = classes
        super().__init__(fr_class)

    def __call__(self, value: str) -> int:
        return self.classes.index(value)



import json

import pandas as pd
from PIL import Image

from dvcx.lib.dataset import Dataset
from dvcx.lib.feature import ShallowFeature
from dvcx.lib.feature_udf import FeatureAggregator
from dvcx.lib.file import File, FileInfo
from dvcx.query.schema import C
from dvcx.sql.functions import path


class BBox(ShallowFeature):
    x_min: int
    x_max: int
    y_min: int
    y_max: int


class OpenImageDetect(FeatureAggregator):
    def __init__(self):
        super().__init__(File, [FileInfo, BBox])

    def process(self, args):
        if len(args) != 2:
            raise ValueError("Group jpg-json mismatch")

        stream_jpg = args[0]
        stream_json = args[1]
        if args[0].get_file_ext() != "jpg":
            stream_jpg, stream_json = stream_json, stream_jpg

        with stream_jpg.open() as fd:
            img = Image.open(fd)

        with stream_json.open() as stream_json:
            detections = json.load(stream_json).get("detections", [])

        for i, detect in enumerate(detections):
            bbox = BBox(
                x_min=int(detect["XMin"] * img.width),
                x_max=int(detect["XMax"] * img.width),
                y_min=int(detect["YMin"] * img.height),
                y_max=int(detect["YMax"] * img.height),
            )

            fstream = FileInfo(
                name=f"detect_{i}",
                source=source,
                parent=f"{stream_jpg.parent}/{stream_jpg.name}",
                version=stream_jpg.version,
                etag=f"{stream_jpg.etag}_{stream_jpg.etag}",
            )

            yield fstream, bbox


source = "s3://ldb-public/remote/data-lakes/open-images-v6-test-200"

ds = (
    Dataset(source, anon=True)
    .filter(C.name.glob("*.jpg") | C.name.glob("*.json"))
    .aggregate(
        OpenImageDetect(),
        partition_by=path.file_stem(C.name),
    )
)

with pd.option_context("display.max_columns", None):
    df = ds.limit(10).to_pandas()
    print(df)



import io
from datetime import datetime

import attrs
import lz4.frame
import pandas as pd
import pytest

from dvcx.dataset import Status as DatasetStatus
from dvcx.error import DVCXError
from dvcx.node import DirType
from tests.data import ENTRIES
from tests.utils import assert_row_names, skip_if_not_sqlite


@pytest.fixture
def dog_entries():
    return [
        attrs.asdict(e) for e in ENTRIES if e.name.startswith("dog") and not e.is_dir
    ]


@pytest.fixture
def dog_entries_parquet_lz4(dog_entries) -> bytes:
    """
    Returns dogs entries in lz4 compressed parquet format
    """

    def _adapt_row(row):
        """
        Adjusting row values to match remote response
        """
        adapted = {}
        for k, v in row.items():
            if isinstance(v, str):
                adapted[k] = v.encode("utf-8")
            elif isinstance(v, datetime):
                adapted[k] = v.timestamp()
            elif v is None:
                adapted[k] = b""
            else:
                adapted[k] = v

        adapted["id"] = 1
        adapted["vtype"] = b""
        adapted["location"] = b""
        adapted["source"] = b"s3://dogs"
        adapted["dir_type"] = DirType.FILE
        adapted["random"] = 1
        return adapted

    dog_entries = [_adapt_row(e) for e in dog_entries]
    df = pd.DataFrame.from_records(dog_entries)
    buffer = io.BytesIO()
    df.to_parquet(buffer, engine="auto")

    return lz4.frame.compress(buffer.getvalue())


@pytest.fixture
def custom_column_types():
    return {
        "id": {"type": "UInt64"},
        "vtype": {"type": "String"},
        "dir_type": {"type": "Int32"},
        "parent_id": {"type": "Int64"},
        "parent": {"type": "String"},
        "name": {"type": "String"},
        "etag": {"type": "String"},
        "version": {"type": "String"},
        "is_latest": {"type": "Boolean"},
        "last_modified": {"type": "DateTime"},
        "size": {"type": "Int64"},
        "owner_name": {"type": "String"},
        "owner_id": {"type": "String"},
        "random": {"type": "Int64"},
        "location": {"type": "String"},
        "source": {"type": "String"},
    }


@pytest.fixture
def remote_dataset_version(custom_column_types, dataset_rows):
    return {
        "id": 1,
        "dataset_id": 1,
        "version": 1,
        "status": 4,
        "created_at": "2024-02-23T10:42:31.842944+00:00",
        "finished_at": "2024-02-23T10:43:31.842944+00:00",
        "error_message": "",
        "error_stack": "",
        "num_objects": 5,
        "size": 1073741824,
        "preview": [dataset_row.to_preview() for dataset_row in dataset_rows],
        "script_output": "",
        "custom_column_types": custom_column_types,
        "sources": "",
        "query_script": (
            'from dvcx.query import DatasetQuery\nDatasetQuery(path="s3://ldb-public")',
        ),
        "created_by_id": 1,
    }


@pytest.fixture
def remote_dataset(remote_dataset_version, custom_column_types):
    return {
        "id": 1,
        "name": "remote",
        "description": "",
        "labels": [],
        "shadow": False,
        "custom_column_types": custom_column_types,
        "status": 4,
        "created_at": "2024-02-23T10:42:31.842944+00:00",
        "finished_at": "2024-02-23T10:43:31.842944+00:00",
        "error_message": "",
        "error_stack": "",
        "script_output": "",
        "job_id": "f74ec414-58b7-437d-81c5-d41e5365abba",
        "sources": "",
        "query_script": "",
        "team_id": 1,
        "warehouse_id": None,
        "created_by_id": 1,
        "versions": [remote_dataset_version],
    }


@pytest.fixture
def remote_config():
    return {
        "url": "http://localhost:8111/api/dvcx",
        "username": "dvcx",
        "token": "isat_1LZKasZwyM46eHk6NHZZh4VCbHPRKUlQaLnYUE1bXb2U8Il0U",
    }


@pytest.mark.parametrize("cloud_type, version_aware", [("s3", False)], indirect=True)
@pytest.mark.parametrize("dataset_uri", ["ds://dogs@v1", "ds://dogs"])
def test_pull_dataset_success(
    requests_mock,
    cloud_test_catalog,
    remote_config,
    remote_dataset,
    dog_entries_parquet_lz4,
    dataset_uri,
):
    skip_if_not_sqlite()
    data_url = (
        "https://studio-blobvault.s3.amazonaws.com/dvcx_ds_export_1_0.parquet.lz4"
    )
    requests_mock.post(f'{remote_config["url"]}/dataset-info', json=remote_dataset)
    requests_mock.post(
        f'{remote_config["url"]}/dataset-stats',
        json={"num_objects": 5, "size": 1000},
    )
    requests_mock.post(f'{remote_config["url"]}/dataset-export', json=[data_url])
    requests_mock.post(
        f'{remote_config["url"]}/dataset-export-status',
        json={"status": "completed"},
    )
    requests_mock.get(data_url, content=dog_entries_parquet_lz4)
    catalog = cloud_test_catalog.catalog

    catalog.pull_dataset(dataset_uri, no_cp=True, remote_config=remote_config)
    # trying to pull multiple times as it should work
    catalog.pull_dataset(dataset_uri, no_cp=True, remote_config=remote_config)

    dataset = catalog.get_dataset("dogs")
    assert dataset.versions_values == [1]
    assert dataset.status == DatasetStatus.COMPLETE
    assert dataset.created_at
    assert dataset.finished_at
    assert dataset.custom_column_types
    dataset_version = dataset.get_version(1)
    assert dataset_version.status == DatasetStatus.COMPLETE
    assert dataset_version.created_at
    assert dataset_version.finished_at
    assert dataset_version.custom_column_types
    assert dataset_version.num_objects == 4
    assert dataset_version.size == 15

    assert_row_names(
        catalog,
        dataset,
        1,
        {
            "dog1",
            "dog2",
            "dog3",
            "dog4",
        },
    )


@pytest.mark.parametrize("cloud_type, version_aware", [("s3", False)], indirect=True)
def test_pull_dataset_wrong_dataset_uri_format(
    requests_mock,
    cloud_test_catalog,
    remote_config,
    remote_dataset,
    dog_entries_parquet_lz4,
):
    skip_if_not_sqlite()
    catalog = cloud_test_catalog.catalog

    with pytest.raises(DVCXError) as exc_info:
        catalog.pull_dataset("wrong", no_cp=True, remote_config=remote_config)
    assert str(exc_info.value) == "Error when parsing dataset uri"


@pytest.mark.parametrize("cloud_type, version_aware", [("s3", False)], indirect=True)
def test_pull_dataset_wrong_version(
    requests_mock,
    cloud_test_catalog,
    remote_config,
    remote_dataset,
):
    skip_if_not_sqlite()
    requests_mock.post(
        f'{remote_config["url"]}/dataset-info',
        json=remote_dataset,
    )
    catalog = cloud_test_catalog.catalog

    with pytest.raises(DVCXError) as exc_info:
        catalog.pull_dataset("ds://dogs@v5", no_cp=True, remote_config=remote_config)
    assert str(exc_info.value) == "Dataset dogs doesn't have version 5 on server"


@pytest.mark.parametrize("cloud_type, version_aware", [("s3", False)], indirect=True)
def test_pull_dataset_not_found_in_remote(
    requests_mock,
    cloud_test_catalog,
    remote_config,
    remote_dataset,
):
    skip_if_not_sqlite()
    requests_mock.post(
        f'{remote_config["url"]}/dataset-info',
        status_code=404,
        json={"message": "Dataset not found"},
    )
    catalog = cloud_test_catalog.catalog

    with pytest.raises(DVCXError) as exc_info:
        catalog.pull_dataset("ds://dogs@v1", no_cp=True, remote_config=remote_config)
    assert str(exc_info.value) == "Error from server: Dataset not found"


@pytest.mark.parametrize("cloud_type, version_aware", [("s3", False)], indirect=True)
def test_pull_dataset_error_on_fetching_stats(
    requests_mock,
    cloud_test_catalog,
    remote_config,
    remote_dataset,
):
    skip_if_not_sqlite()
    requests_mock.post(
        f'{remote_config["url"]}/dataset-info',
        json=remote_dataset,
    )
    requests_mock.post(
        f'{remote_config["url"]}/dataset-stats',
        status_code=400,
        json={"message": "Internal error"},
    )
    catalog = cloud_test_catalog.catalog

    with pytest.raises(DVCXError) as exc_info:
        catalog.pull_dataset("ds://dogs@v1", no_cp=True, remote_config=remote_config)
    assert str(exc_info.value) == "Error from server: Internal error"


@pytest.mark.parametrize("cloud_type, version_aware", [("s3", False)], indirect=True)
@pytest.mark.parametrize("export_status", ["failed", "removed"])
def test_pull_dataset_exporting_dataset_failed_in_remote(
    requests_mock,
    cloud_test_catalog,
    remote_config,
    remote_dataset,
    export_status,
):
    skip_if_not_sqlite()
    data_url = (
        "https://studio-blobvault.s3.amazonaws.com/dvcx_ds_export_1_0.parquet.lz4"
    )
    requests_mock.post(f'{remote_config["url"]}/dataset-info', json=remote_dataset)
    requests_mock.post(
        f'{remote_config["url"]}/dataset-stats',
        json={"num_objects": 5, "size": 1000},
    )
    requests_mock.post(f'{remote_config["url"]}/dataset-export', json=[data_url])
    requests_mock.post(
        f'{remote_config["url"]}/dataset-export-status',
        json={"status": export_status},
    )

    catalog = cloud_test_catalog.catalog

    with pytest.raises(DVCXError) as exc_info:
        catalog.pull_dataset("ds://dogs@v1", no_cp=True, remote_config=remote_config)
    assert str(exc_info.value) == (
        f"Error from server: Dataset export {export_status} in Studio"
    )


@pytest.mark.parametrize("cloud_type, version_aware", [("s3", False)], indirect=True)
def test_pull_dataset_empty_parquet(
    requests_mock,
    cloud_test_catalog,
    remote_config,
    remote_dataset,
    dog_entries_parquet_lz4,
):
    skip_if_not_sqlite()
    data_url = (
        "https://studio-blobvault.s3.amazonaws.com/dvcx_ds_export_1_0.parquet.lz4"
    )
    requests_mock.post(f'{remote_config["url"]}/dataset-info', json=remote_dataset)
    requests_mock.post(
        f'{remote_config["url"]}/dataset-stats',
        json={"num_objects": 5, "size": 1000},
    )
    requests_mock.post(f'{remote_config["url"]}/dataset-export', json=[data_url])
    requests_mock.post(
        f'{remote_config["url"]}/dataset-export-status',
        json={"status": "completed"},
    )
    requests_mock.get(data_url, content=b"")
    catalog = cloud_test_catalog.catalog

    with pytest.raises(RuntimeError):
        catalog.pull_dataset("ds://dogs@v1", no_cp=True, remote_config=remote_config)



import inspect
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Callable, Optional, Union

import sqlalchemy
from sqlalchemy.sql.elements import ColumnElement

from dvcx.lib.feature import Feature
from dvcx.lib.feature_types import FeatureLike, FeatureTypes
from dvcx.lib.feature_udf import FeatureConverter
from dvcx.lib.reader import FeatureReader
from dvcx.lib.udf import (
    Aggregator,
    BatchMapper,
    Generator,
    GroupMapper,
    Mapper,
    UDFBase,
)
from dvcx.query.dataset import (
    DatasetQuery,
    JoinPredicateType,
    PartitionByType,
    detach,
)
from dvcx.query.schema import Column

if TYPE_CHECKING:
    from typing_extensions import Self

C = Column


class Dataset(DatasetQuery):
    """DVCX DataFrame is a 2-dimensional table of rows and columns designed to work
    with AI datasets - datasets of unstructured data such as images, video audio, text,
    etc.

    Each row in the DataFrame acts as a reference to a file (or specific elements within
    a file) and is associated with various file parameters, represented as columns.
    These columns typically include labels, embeddings, or auto-generated labels,
    providing a comprehensive view of the file's attributes.
    """

    def apply(self, func, *args, **kwargs):
        return func(self, *args, **kwargs)

    def map(
        self,
        udf: Union[Callable, UDFBase],
        params=None,
        output=None,
        parallel: Optional[int] = None,
        workers: Union[bool, int] = False,
        min_task_size: Optional[int] = None,
        cache: bool = False,
    ) -> "Self":
        """Create a new dataframe by applying a function to each row to create new
        columns. The function should return a list of new column values corresponding
        to column names defined in the `output` parameter.

        Input-output relationship: 1:1

        Parameters:

        `udf` (Callable or UDFBase): Function or class applied to each row.
        `params`: List of column names used as input for the function. Default
                is taken from udf signature. `Stream()` (file object) can be included.
        `output`: Dictionary defining new columns and their corresponding data types.
                The order of the keys is important because it sets the format for the
                function's output.
        `parallel` (int): Enables parallel processing. Set >1 for multiple threads,
                `True` for using all CPU threads. Default is single-threaded.
        `workers` (int): Enables distributed processing if set >1. `True` means use all
                allowed workers. Only for dvc.ai platform setup.
        `min_task_size` (int): Minimum tasks per batch for distributed processing. Only
                for platform setup.
        `cache` (bool): If True, saves `Stream()`/file objects in local cache for future
                usage. Default is False.

        Returns:

        `DataFrame`: New dataframe with original rows but additional columns as
                specified in `output`.

        Example:
        >>> df.map(lambda name: name[:-4] + ".json", output={"json_file": String}) \
        >>>     .save("new_dataset")
        """

        self._validate_args("map()", parallel, workers, min_task_size)

        if output:
            output = {k: Feature._convert_type(v) for k, v in output.items()}

        udf_obj = self._udf_to_obj(udf, Mapper, "map()", params, output, cache=cache)
        return self.add_signals(
            udf_obj.to_udf_wrapper(),
            parallel=parallel,
            workers=workers,
            min_task_size=min_task_size,
            cache=cache,
        )

    def generate(  # type: ignore[override]
        self,
        udf: Union[Callable, UDFBase],
        params=None,
        output=None,
        parallel: Optional[int] = None,
        workers: Union[bool, int] = False,
        min_task_size: Optional[int] = None,
        cache: bool = False,
    ) -> "Self":
        """Create a new dataframe by applying a function to each row to generate a new
        set of rows and columns.

        Input-output relationship: 1:N

        This method is similar to `map()`, uses the same list of parameters, but with
        two key differences:
        1. It produces multiple rows for each input row. This means several rows can be
           generated from a single file record (like extracting multiple files from a
           single archive).
        2. The function should return the complete set of columns, including both file
           columns and newly generated ones, not just the new columns.

        The function returns a nested list, where each inner list represents a set of
        column values (the same as `map()`).
        """
        self._validate_args("generate()", parallel, workers, min_task_size)

        udf_obj = self._udf_to_obj(
            udf, Generator, "generate()", params, output, cache=cache
        )
        return DatasetQuery.generate(
            self,
            udf_obj.to_udf_wrapper(),
            parallel=parallel,
            workers=workers,
            min_task_size=min_task_size,
            cache=cache,
        )

    def aggregate(
        self,
        udf: Union[Callable, UDFBase],
        params=None,
        output=None,
        partition_by: Optional[PartitionByType] = None,
        parallel: Optional[int] = None,
        workers: Union[bool, int] = False,
        min_task_size: Optional[int] = None,
        batch=1,
        cache: bool = False,
    ) -> "Self":
        """Create a new dataframe by aggregating a group of rows using a specified
        function, resulting in a new set of rows and columns.

        Input-output relationship: N:M

        This method bears similarity to `generate()`, employing a comparable set of
        parameters, yet differs in two crucial aspects:
        1. The `partition_by` parameter: This specifies the column name or a list of
           column names that determine the grouping criteria for aggregation.
        2. Group-based UDF function input: Instead of individual rows, the function
           receives a list all rows within each group defined by `partition_by`.
        """
        self._validate_args("aggregate()", parallel, workers, min_task_size)

        udf_obj = self._udf_to_obj(
            udf, Aggregator, "aggregate()", params, output, batch, cache=cache
        )
        return DatasetQuery.generate(
            self,
            udf_obj.to_udf_wrapper(),
            partition_by=partition_by,
            parallel=parallel,
            workers=workers,
            min_task_size=min_task_size,
            cache=cache,
        )

    def batch_map(
        self,
        udf: Union[Callable, UDFBase],
        params=None,
        output=None,
        parallel: Optional[int] = None,
        workers: Union[bool, int] = False,
        min_task_size: Optional[int] = None,
        batch=1000,
        cache: bool = False,
    ) -> "Self":
        """This is a batch version of map(). It accepts the same parameters plus an
        additional parameter:

        `batch` (int), which sets the batch size. The default batch size is 1000.

        Input-output relationship: N:N
        """
        self._validate_args("map()", parallel, workers, min_task_size)

        udf_obj = self._udf_to_obj(
            udf, BatchMapper, "batch_map()", params, output, batch, cache=cache
        )
        return self.add_signals(
            udf_obj.to_udf_wrapper(),
            parallel=parallel,
            workers=workers,
            min_task_size=min_task_size,
            cache=cache,
        )

    def group_map(
        self,
        udf: Union[Callable, UDFBase],
        params=None,
        output=None,
        partition_by: Optional[PartitionByType] = None,
        parallel: Optional[int] = None,
        workers: Union[bool, int] = False,
        min_task_size: Optional[int] = None,
        cache: bool = False,
    ) -> "Self":
        """Warning: experimental functionality!
        Create a new dataframe by applying a function to group of rows to create new
        columns. The function should return a nested list of new column values
        corresponding to columns defined in the `output` parameter.

        Input-output relationship: N:N

        This method is similar to `map()`, employing a comparable set of
        parameters, yet differs in two crucial aspects:
        1. The `partition_by` parameter: This specifies the column name or a list of
           column names that determine the grouping criteria.
        2. Group-based UDF function input: Instead of individual rows, the function
           receives a list all rows within each group defined by `partition_by`.
        3. Group-based UDF function output: Instead of returning a list of new columns
           values, the function returns nested list of the new columns. Each sub-list
           of the columns should correspond to input row in a given order.
        """
        self._validate_args(
            "group_map()", parallel, workers, min_task_size, partition_by
        )

        udf_obj = self._udf_to_obj(
            udf, GroupMapper, "group_map()", params, output, cache=cache
        )
        return self.add_signals(
            udf_obj.to_udf_wrapper(),
            parallel=parallel,
            workers=workers,
            min_task_size=min_task_size,
            partition_by=partition_by,
            cache=cache,
        )

    def _udf_to_obj(
        self,
        udf,
        target_class: type[UDFBase],
        name: str,
        params=None,
        output=None,
        batch: int = 1,
        cache: bool = False,
    ) -> UDFBase:
        if isinstance(udf, UDFBase):
            if not isinstance(udf, target_class):
                cls_name = target_class.__name__
                raise TypeError(
                    f"{name}: expected an instance derived from {cls_name}"
                    f", but received {udf.name}"
                )
            if params:
                raise ValueError(
                    f"params for BaseUDF class {udf.name} cannot be overwritten"
                )
            if output:
                raise ValueError(
                    f"output for BaseUDF class {udf.name} cannot be overwritten"
                )

            if isinstance(udf, UDFBase):
                udf.set_catalog(self.catalog)

                if cache:
                    udf.enable_caching()

            return udf

        if inspect.isfunction(udf):
            return target_class.from_func(udf, params, output, batch)

        if isinstance(udf, type):
            raise TypeError(
                f"{name} error: The class '{udf}' needs to be instantiated"
                f" as an object before you can use it as UDF"
            )

        if not callable(udf):
            raise TypeError(f"{name} error: instance {udf} must be callable for UDF")

        return target_class.from_func(udf, params, output, batch)

    def _validate_args(
        self, name, parallel, workers, min_task_size=None, partition_by=None
    ) -> None:
        msg = None
        if not isinstance(parallel, int) and parallel is not None:
            msg = (
                f"'parallel' argument must be int or None"
                f", {parallel.__class__.__name__} was given"
            )
        elif not isinstance(workers, bool) and not isinstance(workers, int):
            msg = (
                f"'workers' argument must be int or bool"
                f", {workers.__class__.__name__} was given"
            )
        elif min_task_size is not None and not isinstance(min_task_size, int):
            msg = (
                f"'min_task_size' argument must be int or None"
                f", {min_task_size.__class__.__name__} was given"
            )
        elif (
            partition_by is not None
            and not isinstance(partition_by, ColumnElement)
            and not isinstance(partition_by, Sequence)
        ):
            msg = (
                f"'partition_by' argument must be PartitionByType or None"
                f", {partition_by.__class__.__name__} was given"
            )

        if msg:
            raise TypeError(f"Dataset {name} error: {msg}")

    def _args_to_columns(self, *features):
        uniq_columns = {}
        for fr in features:
            uniq_columns = uniq_columns | dict(self._feature_to_column_name_types(fr))

        return [C(name, typ) for name, typ in uniq_columns.items()]

    @classmethod
    def _feature_to_column_name_types(cls, fr: FeatureLike) -> list[tuple[str, type]]:
        if isinstance(fr, str):
            return [(fr, str)]

        if isinstance(fr, C):
            return [(fr.name, type(fr.type))]

        if isinstance(fr, FeatureReader):
            return cls._feature_to_column_name_types(fr.fr_class)

        if not issubclass(fr, Feature):
            raise TypeError(f"feature or column '{fr}' has a wrong type '{type(fr)}'")

        return fr._to_udf_spec()

    def _extend_features(self, method_name, *args):
        super_func = getattr(super(), method_name)

        columns = self._args_to_columns(*args)
        return super_func(*columns)

    @detach
    def select(self, *args, **kwargs) -> "Self":
        return self._extend_features("select", *args)

    @detach
    def select_except(self, *args) -> "Self":
        return self._extend_features("select_except", *args)

    def get_values(self, *fr_classes: FeatureLike) -> Iterator[Sequence]:
        """
        Iterate over dataset, getting feature values and applying reader calls.
        """
        # Get Feature classes and calls for each FeatureReader
        fr_classes_only, callers = zip(
            *[
                (f.fr_class, f) if isinstance(f, FeatureReader) else (f, lambda x: x)
                for f in fr_classes
            ]
        )
        for features in self.iterate(*fr_classes_only):
            yield [
                call(fr._get_value_with_check()) for fr, call in zip(features, callers)
            ]

    def iterate(self, *fr_classes: FeatureLike) -> Iterator[Sequence[Feature]]:
        fr_classes_only = FeatureTypes._to_features(*fr_classes)
        ds = self.select(*fr_classes_only)

        with ds.as_iterable() as rows_iter:
            params = FeatureConverter.get_flattened_params(fr_classes_only)
            for row in rows_iter:
                yield from FeatureConverter.deserialize(
                    [row], params, fr_classes_only, self.catalog, True
                )

    def to_pytorch(self, *fr_classes: FeatureLike, **kwargs):
        try:
            import torch  # noqa: F401
        except ImportError as exc:
            raise ImportError(
                "Missing required dependency 'torch' for Dataset.to_pytorch()"
            ) from exc
        from dvcx.lib.pytorch import PytorchDataset

        if self.attached:
            ds = self
        else:
            ds = self.save()
        assert ds.name is not None  # for mypy
        return PytorchDataset(
            fr_classes, ds.name, ds.version, catalog=self.catalog, **kwargs
        )

    @detach
    def merge(
        self,
        right_ds: "Dataset",
        on: Union[JoinPredicateType, Sequence[JoinPredicateType]],
        right_on: Union[JoinPredicateType, Sequence[JoinPredicateType], None] = None,
        inner=False,
        rname="{name}_right",
    ) -> "Self":
        if on is None:
            raise ValueError("'on' must be specified in merge()")

        list_on = [on] if not isinstance(on, (list, tuple)) else on
        on_columns = [cols for item in list_on for cols in self._args_to_columns(item)]

        if right_on is not None:
            list_right_on = (
                [right_on] if not isinstance(right_on, (list, tuple)) else right_on
            )
            right_on_columns = [
                cols for item in list_right_on for cols in self._args_to_columns(item)
            ]

            if len(right_on_columns) != len(on_columns):
                raise ValueError("'on' and 'right_on' must have the same length'")
        else:
            right_on_columns = on_columns

        ops = [
            self.c(left) == right_ds.c(right)
            for left, right in zip(on_columns, right_on_columns)
        ]

        return self.join(right_ds, sqlalchemy.and_(*ops), inner, rname)



import typing
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import dataclass
from functools import WRAPPER_ASSIGNMENTS
from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Union,
)

from fsspec.callbacks import DEFAULT_CALLBACK, Callback

from dvcx.dataset import DatasetRow

from .batch import Batch, BatchingStrategy, DatasetRowsBatch, NoBatching, Partition
from .schema import (
    UDFParameter,
    UDFParamSpec,
    normalize_param,
)

if TYPE_CHECKING:
    from dvcx.catalog import Catalog

    from .batch import BatchingResult

ColumnType = Any


# Specification for the output of a UDF
UDFOutputSpec = typing.Mapping[str, ColumnType]

# Result type when calling the UDF wrapper around the actual
# Python function / class implementing it.
UDFResult = dict[str, Any]


@dataclass
class UDFProperties:
    """Container for basic UDF properties."""

    params: list[UDFParameter]
    output: UDFOutputSpec
    batch: int = 1

    def get_batching(self, use_partitioning: bool = False) -> BatchingStrategy:
        if use_partitioning:
            return Partition()
        if self.batch == 1:
            return NoBatching()
        if self.batch > 1:
            return Batch(self.batch)
        raise ValueError(f"invalid batch size {self.batch}")

    def signal_names(self) -> Iterable[str]:
        return self.output.keys()

    def parameter_parser(self) -> Callable:
        """Generate a parameter list from a dataset row."""

        def plist(catalog: "Catalog", row: "DatasetRow", **kwargs) -> list:
            return [p.get_value(catalog, row, **kwargs) for p in self.params]

        return plist


def udf(
    params: Sequence[UDFParamSpec],
    output: UDFOutputSpec,
    *,
    method: Optional[str] = None,  # only used for class-based UDFs
    batch: int = 1,
):
    """
    Decorate a function or a class to be used as a UDF.

    The decorator expects both the outputs and inputs of the UDF to be specified.
    The outputs are defined as a collection of tuples containing the signal name
    and type.
    Parameters are defined as a list of column objects (e.g. C.name).
    Optionally, UDFs can be run on batches of rows to improve performance, this
    is determined by the 'batch' parameter. When operating on batches of inputs,
    the UDF function will be called with a single argument - a list
    of tuples containing inputs (e.g. ((input1_a, input1_b), (input2_a, input2b))).
    """
    if isinstance(params, str):
        params = (params,)
    if not isinstance(output, Mapping):
        raise TypeError(f"'output' must be a mapping, got {type(output).__name__}")

    properties = UDFProperties([normalize_param(p) for p in params], output, batch)

    def decorator(udf_base: Union[Callable, type]):
        if isclass(udf_base):
            return UDFClassWrapper(udf_base, properties, method=method)
        if callable(udf_base):
            return UDFWrapper(udf_base, properties)

    return decorator


class UDFBase:
    """A base class for implementing stateful UDFs."""

    def __init__(
        self,
        func: Callable,
        properties: UDFProperties,
    ):
        self.func = func
        self.properties = properties
        self.signal_names = properties.signal_names()
        self.parameter_parser = properties.parameter_parser()
        self.output = properties.output

    def __call__(
        self,
        catalog: "Catalog",
        param: "BatchingResult",
        is_generator: bool = False,
        cache: bool = False,
        cb: Callback = DEFAULT_CALLBACK,
    ) -> Iterable[UDFResult]:
        if isinstance(param, DatasetRowsBatch):
            udf_inputs = [
                self.parameter_parser(catalog, row, cache=cache, cb=cb)
                for row in param.rows
            ]
            udf_outputs = self.func(udf_inputs)
            return self._process_results(param.rows, udf_outputs, is_generator)
        if isinstance(param, DatasetRow):
            udf_inputs = self.parameter_parser(catalog, param, cache=cache, cb=cb)
            udf_outputs = self.func(*udf_inputs)
            if not is_generator:
                # udf_outputs is generator already if is_generator=True
                udf_outputs = [udf_outputs]
            return self._process_results([param], udf_outputs, is_generator)
        raise ValueError(f"unexpected UDF parameter {param}")

    def _process_results(
        self,
        rows: Sequence["DatasetRow"],
        results: Sequence[Sequence[Any]],
        is_generator=False,
    ) -> Iterable[UDFResult]:
        """Create a list of dictionaries representing UDF results."""

        # outputting rows
        if is_generator:
            # each row in results is a tuple of column values
            return (dict(zip(self.signal_names, row)) for row in results)

        # outputting signals
        row_ids = [row.id for row in rows]
        return [
            dict(id=row_id, **dict(zip(self.signal_names, signals)))
            for row_id, signals in zip(row_ids, results)
            if signals is not None  # skip rows with no output
        ]


class UDFClassWrapper:
    """
    A wrapper for class-based (stateful) UDFs.
    """

    def __init__(
        self,
        udf_class: type,
        properties: UDFProperties,
        method: Optional[str] = None,
    ):
        self.udf_class = udf_class
        self.udf_method = method
        self.properties = properties
        self.output = properties.output

    def __call__(self, *args, **kwargs) -> "UDFFactory":
        return UDFFactory(
            self.udf_class,
            args,
            kwargs,
            self.properties,
            self.udf_method,
        )


class UDFWrapper(UDFBase):
    """A wrapper class for function UDFs to be used in custom signal generation."""

    def __init__(
        self,
        func: Callable,
        properties: UDFProperties,
    ):
        super().__init__(func, properties)
        # This emulates the behavior of functools.wraps for a class decorator
        for attr in WRAPPER_ASSIGNMENTS:
            if hasattr(func, attr):
                setattr(self, attr, getattr(func, attr))

    # This emulates the behavior of functools.wraps for a class decorator
    def __repr__(self):
        return repr(self.func)


class UDFFactory:
    """
    A wrapper for late instantiation of UDF classes, primarily for use in parallelized
    execution.
    """

    def __init__(
        self,
        udf_class: type,
        args,
        kwargs,
        properties: UDFProperties,
        method: Optional[str] = None,
    ):
        self.udf_class = udf_class
        self.udf_method = method
        self.args = args
        self.kwargs = kwargs
        self.properties = properties
        self.output = properties.output

    def __call__(self) -> UDFWrapper:
        udf_func = self.udf_class(*self.args, **self.kwargs)
        if self.udf_method:
            udf_func = getattr(udf_func, self.udf_method)

        return UDFWrapper(udf_func, self.properties)


UDFType = Union[UDFBase, UDFFactory]



import open_clip
import pytest
import torch

from dvcx.lib.feature_types import FeatureTypes
from dvcx.lib.file import TextFile
from dvcx.lib.text import TextReader, convert_text


@pytest.mark.parametrize("tokenizer_model", ["ViT-B-32", "hf-hub:timm/ViT-B-16-SigLIP"])
def test_convert_text(tmp_path, tokenizer_model):
    text = "thisismytext"
    tokenizer = open_clip.get_tokenizer(tokenizer_model)
    tokenizer_kwargs = {"context_length": 100}
    converted_text = convert_text(
        text, tokenizer=tokenizer, tokenizer_kwargs=tokenizer_kwargs
    )
    assert isinstance(converted_text, torch.Tensor)
    assert converted_text.size() == (100,)


def test_text_reader_column(mocker):
    text = "mytext"
    TextFeature = FeatureTypes.column_class("text")  # noqa: N806
    text_feature = TextFeature(text=text)

    mock_tokenizer = mocker.Mock()
    kwargs = {
        "tokenizer": mock_tokenizer,
        "tokenizer_kwargs": {"some": "arg"},
        "open_clip_model": None,
    }
    reader = TextReader("text", **kwargs)

    convert_text = mocker.patch("dvcx.lib.text.convert_text")
    reader(text_feature.get_value())
    convert_text.assert_called_with(text, **kwargs)


def test_text_reader_file(tmp_path, catalog, mocker):
    file_name = "myfile"
    text = "myText"

    file_path = tmp_path / file_name
    with open(file_path, "w") as fd:
        fd.write(text)

    mock_tokenizer = mocker.Mock()
    kwargs = {
        "tokenizer": mock_tokenizer,
        "tokenizer_kwargs": {"some": "arg"},
        "open_clip_model": None,
    }
    reader = TextReader(**kwargs)
    file = TextFile(name=file_name, source=f"file://{tmp_path}")
    file.set_catalog(catalog)

    convert_text = mocker.patch("dvcx.lib.text.convert_text")
    with open(file_path) as fd:
        file.set_file(fd, caching_enabled=False)
        reader(file.get_value())
        convert_text.assert_called_with(text, **kwargs)



import logging
import os
import sqlite3
from collections.abc import Iterable, Sequence
from contextlib import contextmanager
from functools import wraps
from time import sleep
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Optional,
    Union,
)

import sqlalchemy
from attrs import frozen
from sqlalchemy import MetaData, Table, UniqueConstraint, exists, select
from sqlalchemy.dialects import sqlite
from sqlalchemy.schema import CreateIndex, CreateTable, DropTable
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import bindparam, cast

import dvcx.sql.sqlite
from dvcx.data_storage import AbstractDBMetastore, AbstractWarehouse
from dvcx.data_storage.db_engine import DatabaseEngine
from dvcx.data_storage.id_generator import AbstractDBIDGenerator
from dvcx.data_storage.schema import (
    DefaultSchema,
    convert_rows_custom_column_types,
)
from dvcx.dataset import DatasetRecord
from dvcx.error import DVCXError
from dvcx.sql.sqlite import create_user_defined_sql_functions, sqlite_dialect
from dvcx.sql.types import SQLType
from dvcx.storage import StorageURI
from dvcx.utils import DVCXDir

if TYPE_CHECKING:
    from sqlalchemy.schema import SchemaItem
    from sqlalchemy.sql.elements import ColumnClause, ColumnElement, TextClause
    from sqlalchemy.types import TypeEngine


logger = logging.getLogger("dvcx")

RETRY_START_SEC = 0.01
RETRY_MAX_TIMES = 10
RETRY_FACTOR = 2

Column = Union[str, "ColumnClause[Any]", "TextClause"]

dvcx.sql.sqlite.setup()

quote_schema = sqlite_dialect.identifier_preparer.quote_schema
quote = sqlite_dialect.identifier_preparer.quote


def get_retry_sleep_sec(retry_count: int) -> int:
    return RETRY_START_SEC * (RETRY_FACTOR**retry_count)


def retry_sqlite_locks(func):
    # This retries the database modification in case of concurrent access
    @wraps(func)
    def wrapper(*args, **kwargs):
        exc = None
        for retry_count in range(RETRY_MAX_TIMES):
            try:
                return func(*args, **kwargs)
            except sqlite3.OperationalError as operror:
                exc = operror
                sleep(get_retry_sleep_sec(retry_count))
        raise exc

    return wrapper


@frozen
class SQLiteDatabaseEngine(DatabaseEngine):
    dialect = sqlite_dialect

    db: sqlite3.Connection
    db_file: Optional[str]

    @classmethod
    def from_db_file(cls, db_file: Optional[str] = None) -> "SQLiteDatabaseEngine":
        detect_types = sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES

        try:
            if db_file == ":memory:":
                # Enable multithreaded usage of the same in-memory db
                db = sqlite3.connect(
                    "file::memory:?cache=shared", uri=True, detect_types=detect_types
                )
            else:
                db = sqlite3.connect(
                    db_file or DVCXDir.find().db, detect_types=detect_types
                )
            create_user_defined_sql_functions(db)
            engine = sqlalchemy.create_engine(
                "sqlite+pysqlite:///", creator=lambda: db, future=True
            )

            db.isolation_level = None  # Use autocommit mode
            db.execute("PRAGMA foreign_keys = ON")
            db.execute("PRAGMA cache_size = -102400")  # 100 MiB
            # Enable Write-Ahead Log Journaling
            db.execute("PRAGMA journal_mode = WAL")
            db.execute("PRAGMA synchronous = NORMAL")
            db.execute("PRAGMA case_sensitive_like = ON")
            if os.environ.get("DEBUG_SHOW_SQL_QUERIES"):
                db.set_trace_callback(print)

            return cls(engine, MetaData(), db, db_file)
        except RuntimeError:
            raise DVCXError("Can't connect to SQLite DB") from None

    def clone(self) -> "SQLiteDatabaseEngine":
        """Clones DatabaseEngine implementation."""
        return SQLiteDatabaseEngine.from_db_file(self.db_file)

    def clone_params(self) -> tuple[Callable[..., Any], list[Any], dict[str, Any]]:
        """
        Returns the function, args, and kwargs needed to instantiate a cloned copy
        of this DatabaseEngine implementation, for use in separate processes
        or machines.
        """
        return (
            SQLiteDatabaseEngine.from_db_file,
            [self.db_file],
            {},
        )

    @retry_sqlite_locks
    def execute(
        self,
        query,
        cursor: Optional[sqlite3.Cursor] = None,
        conn=None,
    ) -> sqlite3.Cursor:
        if cursor is not None:
            result = cursor.execute(*self.compile_to_args(query))
        elif conn is not None:
            result = conn.execute(*self.compile_to_args(query))
        else:
            result = self.db.execute(*self.compile_to_args(query))
        if isinstance(query, CreateTable) and query.element.indexes:
            for index in query.element.indexes:
                self.execute(CreateIndex(index, if_not_exists=True), cursor=cursor)
        return result

    @retry_sqlite_locks
    def executemany(
        self, query, params, cursor: Optional[sqlite3.Cursor] = None
    ) -> sqlite3.Cursor:
        if cursor:
            return cursor.executemany(self.compile(query).string, params)
        return self.db.executemany(self.compile(query).string, params)

    def execute_str(self, sql: str, parameters=None) -> sqlite3.Cursor:
        if parameters is None:
            return self.db.execute(sql)
        return self.db.execute(sql, parameters)

    def insert_dataframe(self, table_name: str, df) -> int:
        return df.to_sql(table_name, self.db, if_exists="append", index=False)

    def cursor(self, factory=None):
        if factory is None:
            return self.db.cursor()
        return self.db.cursor(factory)

    def close(self) -> None:
        self.db.close()

    @contextmanager
    def transaction(self):
        db = self.db
        with db:
            db.execute("begin")
            yield db

    def has_table(self, name: str) -> bool:
        """
        Return True if a table exists with the given name

        We cannot simply use `inspect(engine).has_table(name)` like the
        parent class does because that will return False for a table
        created during a pending transaction. Instead, we check the
        sqlite_master table.
        """
        query = select(
            exists(
                select(1)
                .select_from(sqlalchemy.table("sqlite_master"))
                .where(
                    (sqlalchemy.column("type") == "table")
                    & (sqlalchemy.column("name") == name)
                )
            )
        )
        return bool(next(self.execute(query))[0])

    def create_table(self, table: "Table", if_not_exists: bool = True) -> None:
        self.execute(CreateTable(table, if_not_exists=if_not_exists))

    def drop_table(self, table: "Table", if_exists: bool = False) -> None:
        self.execute(DropTable(table, if_exists=if_exists))

    def rename_table(self, old_name: str, new_name: str):
        comp_old_name = quote_schema(old_name)
        comp_new_name = quote_schema(new_name)
        self.execute_str(f"ALTER TABLE {comp_old_name} RENAME TO {comp_new_name}")


class SQLiteIDGenerator(AbstractDBIDGenerator):
    _db: "SQLiteDatabaseEngine"

    def __init__(
        self,
        db: Optional["SQLiteDatabaseEngine"] = None,
        table_prefix: Optional[str] = None,
        skip_db_init: bool = False,
        db_file: Optional[str] = None,
    ):
        db = db or SQLiteDatabaseEngine.from_db_file(db_file)

        super().__init__(db, table_prefix, skip_db_init)

    def clone(self) -> "SQLiteIDGenerator":
        """Clones SQLiteIDGenerator implementation."""
        return SQLiteIDGenerator(
            self._db.clone(), self._table_prefix, skip_db_init=True
        )

    def clone_params(self) -> tuple[Callable[..., Any], list[Any], dict[str, Any]]:
        """
        Returns the function, args, and kwargs needed to instantiate a cloned copy
        of this SQLiteIDGenerator implementation, for use in separate processes
        or machines.
        """
        return (
            SQLiteIDGenerator.init_after_clone,
            [],
            {
                "db_clone_params": self._db.clone_params(),
                "table_prefix": self._table_prefix,
            },
        )

    @classmethod
    def init_after_clone(
        cls,
        *,
        db_clone_params: tuple[Callable, list, dict[str, Any]],
        table_prefix: Optional[str] = None,
    ) -> "SQLiteIDGenerator":
        """
        Initializes a new instance of this SQLiteIDGenerator implementation
        using the given parameters, which were obtained from a call to clone_params.
        """
        (db_class, db_args, db_kwargs) = db_clone_params
        return cls(
            db=db_class(*db_args, **db_kwargs),
            table_prefix=table_prefix,
            skip_db_init=True,
        )

    @property
    def db(self) -> "SQLiteDatabaseEngine":
        return self._db

    def init_id(self, uri: str) -> None:
        """Initializes the ID generator for the given URI with zero last_id."""
        self._db.execute(
            sqlite.insert(self._table)
            .values(uri=uri, last_id=0)
            .on_conflict_do_nothing()
        )

    def get_next_ids(self, uri: str, count: int) -> range:
        """Returns a range of IDs for the given URI."""

        # NOTE: we can't use RETURNING clause here because it is only available
        # in sqlalchemy v2, see
        # https://github.com/sqlalchemy/sqlalchemy/issues/6195#issuecomment-1248700677
        # After we upgrade to sqlalchemy v2, we can use the following code,
        # leaving fallback to the current implementation for older versions of SQLite,
        # which is still supported, for example, in Ubuntu 20.04 LTS (Focal Fossa),
        # where SQLite version 3.31.1 is used.

        # sqlite_version = version.parse(sqlite3.sqlite_version)
        # if sqlite_version >= version.parse("3.35.0"):
        #     # RETURNING is supported on SQLite 3.35.0 (2021-03-12) or newer
        #     stmt = (
        #         sqlite.insert(self._table)
        #         .values(uri=uri, last_id=count)
        #         .on_conflict_do_update(
        #             index_elements=["uri"],
        #             set_={"last_id": self._table.c.last_id + count},
        #         )
        #         .returning(self._table.c.last_id)
        #     )
        #     last_id = self._db.execute(stmt).fetchone()[0]
        # else:
        #     (fallback to the current implementation with a transaction)

        # Transactions ensure no concurrency conflicts
        with self._db.transaction() as conn:
            # UPSERT syntax was added to SQLite with version 3.24.0 (2018-06-04).
            stmt_ins = (
                sqlite.insert(self._table)
                .values(uri=uri, last_id=count)
                .on_conflict_do_update(
                    index_elements=["uri"],
                    set_={"last_id": self._table.c.last_id + count},
                )
            )
            self._db.execute(stmt_ins, conn=conn)

            stmt_sel = (
                select(self._table.c.last_id)
                .select_from(self._table)
                .where(self._table.c.uri == uri)
            )
            last_id = self._db.execute(stmt_sel, conn=conn).fetchone()[0]

        return range(last_id - count + 1, last_id + 1)


class SQLiteMetastore(AbstractDBMetastore):
    """
    SQLite Metastore uses SQLite3 for storing indexed data locally.
    This is currently used for the local cli.
    """

    id_generator: "SQLiteIDGenerator"
    db: "SQLiteDatabaseEngine"

    def __init__(
        self,
        id_generator: "SQLiteIDGenerator",
        uri: StorageURI = StorageURI(""),
        partial_id: Optional[int] = None,
        db: Optional["SQLiteDatabaseEngine"] = None,
        db_file: Optional[str] = None,
    ):
        self.schema: DefaultSchema = DefaultSchema()
        super().__init__(id_generator, uri, partial_id)

        # needed for dropping tables in correct order for tests because of
        # foreign keys
        self.default_table_names: list[str] = []

        self.db = db or SQLiteDatabaseEngine.from_db_file(db_file)

        self._init_storage_table()
        self._init_datasets_tables()

    def clone(
        self,
        uri: StorageURI = StorageURI(""),
        partial_id: Optional[int] = None,
        use_new_connection: bool = False,
    ) -> "SQLiteMetastore":
        if not uri:
            if partial_id is not None:
                raise ValueError("if partial_id is used, uri cannot be empty")
            if self.uri:
                uri = self.uri
                if self.partial_id:
                    partial_id = self.partial_id
        return SQLiteMetastore(
            self.id_generator.clone(),
            uri=uri,
            partial_id=partial_id,
            db=self.db.clone(),
        )

    def clone_params(self) -> tuple[Callable[..., Any], list[Any], dict[str, Any]]:
        """
        Returns the class, args, and kwargs needed to instantiate a cloned copy of this
        SQLiteDataStorage implementation, for use in separate processes or machines.
        """
        return (
            SQLiteMetastore.init_after_clone,
            [],
            {
                "id_generator_clone_params": self.id_generator.clone_params(),
                "uri": self.uri,
                "partial_id": self.partial_id,
                "db_clone_params": self.db.clone_params(),
            },
        )

    @classmethod
    def init_after_clone(
        cls,
        *,
        id_generator_clone_params: tuple[Callable, list, dict[str, Any]],
        uri: StorageURI,
        partial_id: Optional[int],
        db_clone_params: tuple[Callable, list, dict[str, Any]],
    ) -> "SQLiteMetastore":
        (
            id_generator_class,
            id_generator_args,
            id_generator_kwargs,
        ) = id_generator_clone_params
        (db_class, db_args, db_kwargs) = db_clone_params
        return cls(
            id_generator=id_generator_class(*id_generator_args, **id_generator_kwargs),
            uri=uri,
            partial_id=partial_id,
            db=db_class(*db_args, **db_kwargs),
        )

    def _init_storage_table(self) -> None:
        """Initialize only tables related to storage, e.g. s3"""
        self.db.create_table(self._storages, if_not_exists=True)
        self.default_table_names.append(self._storages.name)

    def _init_datasets_tables(self) -> None:
        self.db.create_table(self._datasets, if_not_exists=True)
        self.default_table_names.append(self._datasets.name)
        self.db.create_table(self._datasets_versions, if_not_exists=True)
        self.default_table_names.append(self._datasets_versions.name)
        self.db.create_table(self._datasets_dependencies, if_not_exists=True)
        self.default_table_names.append(self._datasets_dependencies.name)

    def init(self, uri: StorageURI, partial_id: int) -> None:
        if not uri:
            raise ValueError("uri for init() cannot be empty")
        partials_table = self._partials_table(uri)
        self.db.create_table(partials_table, if_not_exists=True)

    @classmethod
    def _buckets_columns(cls) -> list["SchemaItem"]:
        """Buckets (storages) table columns."""
        return [*super()._buckets_columns(), UniqueConstraint("uri")]

    @classmethod
    def _datasets_columns(cls) -> list["SchemaItem"]:
        """Datasets table columns."""
        return [*super()._datasets_columns(), UniqueConstraint("name")]

    def _storages_insert(self):
        return sqlite.insert(self._storages)

    def _partials_insert(self):
        return sqlite.insert(self._partials)

    def _datasets_insert(self):
        return sqlite.insert(self._datasets)

    def _datasets_versions_insert(self):
        return sqlite.insert(self._datasets_versions)

    def _datasets_dependencies_insert(self):
        return sqlite.insert(self._datasets_dependencies)

    #
    # Storages
    #

    def mark_storage_not_indexed(self, uri: StorageURI) -> None:
        """
        Mark storage as not indexed.
        This method should be called when storage index is deleted.
        """
        self.db.execute(self._storages_delete().where(self._storages.c.uri == uri))

    #
    # Dataset dependencies
    #

    def _dataset_dependencies_select_columns(self) -> list["SchemaItem"]:
        return [
            self._datasets_dependencies.c.id,
            self._datasets_dependencies.c.dataset_id,
            self._datasets_dependencies.c.dataset_version_id,
            self._datasets_dependencies.c.bucket_id,
            self._datasets_dependencies.c.bucket_version,
            self._datasets.c.name,
            self._datasets.c.created_at,
            self._datasets_versions.c.version,
            self._datasets_versions.c.created_at,
            self._storages.c.uri,
        ]


class SQLiteWarehouse(AbstractWarehouse):
    """
    SQLite Warehouse uses SQLite3 for storing indexed data locally.
    This is currently used for the local cli.
    """

    id_generator: "SQLiteIDGenerator"
    db: "SQLiteDatabaseEngine"

    # Cache for our defined column types to dialect specific TypeEngine relations
    _col_python_type: ClassVar[dict[type, "TypeEngine"]] = {}

    def __init__(
        self,
        id_generator: "SQLiteIDGenerator",
        db: Optional["SQLiteDatabaseEngine"] = None,
        db_file: Optional[str] = None,
    ):
        self.schema: DefaultSchema = DefaultSchema()
        super().__init__(id_generator)

        self.db = db or SQLiteDatabaseEngine.from_db_file(db_file)

    def clone(self, use_new_connection: bool = False) -> "SQLiteWarehouse":
        return SQLiteWarehouse(self.id_generator.clone(), db=self.db.clone())

    def clone_params(self) -> tuple[Callable[..., Any], list[Any], dict[str, Any]]:
        """
        Returns the class, args, and kwargs needed to instantiate a cloned copy of this
        SQLiteDataStorage implementation, for use in separate processes or machines.
        """
        return (
            SQLiteWarehouse.init_after_clone,
            [],
            {
                "id_generator_clone_params": self.id_generator.clone_params(),
                "db_clone_params": self.db.clone_params(),
            },
        )

    @classmethod
    def init_after_clone(
        cls,
        *,
        id_generator_clone_params: tuple[Callable, list, dict[str, Any]],
        db_clone_params: tuple[Callable, list, dict[str, Any]],
    ) -> "SQLiteWarehouse":
        (
            id_generator_class,
            id_generator_args,
            id_generator_kwargs,
        ) = id_generator_clone_params
        (db_class, db_args, db_kwargs) = db_clone_params
        return cls(
            id_generator=id_generator_class(*id_generator_args, **id_generator_kwargs),
            db=db_class(*db_args, **db_kwargs),
        )

    def _reflect_tables(self, filter_tables=None):
        """
        Since some tables are prone to schema extension, meaning we can add
        additional columns to it, we should reflect changes in metadata
        to have the latest columns when dealing with those tables.
        If filter function is defined, it's used to filter out tables to reflect,
        otherwise all tables are reflected
        """
        self.db.metadata.reflect(
            bind=self.db.engine,
            extend_existing=True,
            only=filter_tables,
        )

    def is_ready(self, timeout: Optional[int] = None) -> bool:
        return True

    def create_dataset_rows_table(
        self,
        name: str,
        custom_columns: Sequence["sqlalchemy.Column"] = (),
        if_not_exists: bool = True,
    ) -> Table:
        table = self.schema.dataset_row_cls.new_table(
            name, custom_columns=custom_columns, metadata=self.db.metadata
        )
        self.db.create_table(table, if_not_exists=if_not_exists)
        return table

    def dataset_rows_select(
        self, select_query: sqlalchemy.sql.selectable.Select, **kwargs
    ):
        rows = self.db.execute(select_query, **kwargs)
        yield from convert_rows_custom_column_types(
            select_query.columns, rows, sqlite_dialect
        )

    def get_dataset_sources(
        self, dataset: DatasetRecord, version: int
    ) -> list[StorageURI]:
        dr = self.dataset_rows(dataset, version)
        query = dr.select(dr.c.source).distinct()
        cur = self.db.cursor()
        cur.row_factory = sqlite3.Row  # type: ignore[assignment]

        return [StorageURI(row["source"]) for row in self.db.execute(query, cursor=cur)]

    def merge_dataset_rows(
        self,
        src: DatasetRecord,
        dst: DatasetRecord,
        src_version: int,
        dst_version: int,
    ) -> None:
        dst_empty = False

        if not self.db.has_table(self.dataset_table_name(src.name, src_version)):
            # source table doesn't exist, nothing to do
            return

        src_dr = self.dataset_rows(src, src_version).table

        if not self.db.has_table(self.dataset_table_name(dst.name, dst_version)):
            # destination table doesn't exist, create it
            custom_columns = [
                c
                for c in src_dr.c
                if c.name
                not in [
                    c.name
                    for c in self.schema.dataset_row_cls.dataset_default_columns()
                ]
            ]
            self.create_dataset_rows_table(
                self.dataset_table_name(dst.name, dst_version),
                custom_columns=custom_columns,
            )
            dst_empty = True

        dst_dr = self.dataset_rows(dst, dst_version).table
        merge_fields = [c.name for c in src_dr.c if c.name != "id"]
        select_src = select(*(getattr(src_dr.c, f) for f in merge_fields))

        if dst_empty:
            # we don't need union, but just select from source to destination
            insert_query = sqlite.insert(dst_dr).from_select(merge_fields, select_src)
        else:
            dst_version_latest = None
            # find the previous version of the destination dataset
            dst_previous_versions = [
                v.version
                for v in dst.versions  # type: ignore [union-attr]
                if v.version < dst_version
            ]
            if dst_previous_versions:
                dst_version_latest = max(dst_previous_versions)

            dst_dr_latest = self.dataset_rows(dst, dst_version_latest).table

            select_dst_latest = select(
                *(getattr(dst_dr_latest.c, f) for f in merge_fields)
            )
            union_query = sqlalchemy.union(select_src, select_dst_latest)
            insert_query = (
                sqlite.insert(dst_dr)
                .from_select(merge_fields, union_query)
                .prefix_with("OR IGNORE")
            )

        self.db.execute(insert_query)

    def insert_rows(self, table: Table, rows: Iterable[dict[str, Any]]) -> None:
        rows = list(rows)
        if not rows:
            return
        self.db.executemany(
            table.insert().values({f: bindparam(f) for f in rows[0]}),
            rows,
        )

    def insert_dataset_rows(self, df, dataset: DatasetRecord, version: int) -> int:
        dr = self.dataset_rows(dataset, version)
        return self.db.insert_dataframe(dr.table.name, df)

    def instr(self, source, target) -> "ColumnElement":
        return cast(func.instr(source, target), sqlalchemy.Boolean)

    def get_table(self, name: str) -> sqlalchemy.Table:
        # load table with latest schema to metadata
        self._reflect_tables(filter_tables=lambda t, _: t == name)
        return self.db.metadata.tables[name]

    def python_type(self, col_type: Union["TypeEngine", "SQLType"]) -> Any:
        if isinstance(col_type, SQLType):
            # converting our defined column types to dialect specific TypeEngine
            col_type_cls = type(col_type)
            if col_type_cls not in self._col_python_type:
                self._col_python_type[col_type_cls] = col_type.type_engine(
                    sqlite_dialect
                )
            col_type = self._col_python_type[col_type_cls]

        return col_type.python_type

    def dataset_table_export_file_names(
        self, dataset: DatasetRecord, version: int
    ) -> list[str]:
        raise NotImplementedError("Exporting dataset table not implemented for SQLite")

    def export_dataset_table(
        self,
        bucket_uri: str,
        dataset: DatasetRecord,
        version: int,
        client_config=None,
    ) -> list[str]:
        raise NotImplementedError("Exporting dataset table not implemented for SQLite")



import logging

from dvcx.nodes_thread_pool import NodesThreadPool

logger = logging.getLogger("dvcx")


class NodesFetcher(NodesThreadPool):
    def __init__(self, client, max_threads, cache):
        super().__init__(max_threads)
        self.client = client
        self.cache = cache

    def done_task(self, done):
        for task in done:
            task.result()

    def do_task(self, chunk):
        from fsspec import Callback

        class _CB(Callback):
            def relative_update(_, inc: int = 1):  # noqa: N805
                self.increase_counter(inc)

        for node in chunk:
            uid = node.as_uid(self.client.uri)
            if self.cache.contains(uid):
                self.increase_counter(node.size)
            else:
                self.client.put_in_cache(uid, callback=_CB())



from datetime import datetime
from typing import Any, Optional

import attrs

from dvcx.cache import UniqueId
from dvcx.storage import StorageURI
from dvcx.utils import time_to_str


class DirType:
    FILE = 0
    DIR = 1
    TAR_ARCHIVE = 5


class DirTypeGroup:
    """
    Groups of DirTypes for selecting storage nodes or dataset entries.

    When filtering with FILE and DIR together or alternatively when
    using SUBOBJ_FILE and SUBOBJ_DIR together, we achieve a
    filesystem-compatible view of a storage location. Such a view
    avoids path conflicts and could be downloaded as a directory tree.

    FILE, DIR
      The respective types which appear on the indexed filesystem or
      object store as a file or directory.

    SUBOBJ_FILE, SUBOBJ_DIR
      The respective types that we want to consider to be a file or
      directory when including subobjects which are generated from other
      files. In this case, we treat tar archives as directories so tar
      subobjects (TAR_FILE) can be viewed under the directory tree of
      the parent tar archive.
    """

    FILE = (DirType.FILE, DirType.TAR_ARCHIVE)
    DIR = (DirType.DIR,)
    SUBOBJ_FILE = (DirType.FILE,)
    SUBOBJ_DIR = (DirType.DIR, DirType.TAR_ARCHIVE)


@attrs.define
class Node:
    id: int = 0
    vtype: str = ""
    dir_type: Optional[int] = None
    parent: str = ""
    name: str = ""
    etag: str = ""
    version: Optional[str] = None
    is_latest: bool = True
    last_modified: Optional[datetime] = None
    size: int = 0
    owner_name: str = ""
    owner_id: str = ""
    random: int = -1
    location: Optional[str] = None
    source: StorageURI = StorageURI("")

    @property
    def path(self) -> str:
        return f"{self.parent}/{self.name}" if self.parent else self.name

    @property
    def is_dir(self) -> bool:
        return self.dir_type == DirType.DIR

    @property
    def is_container(self) -> bool:
        return self.dir_type in DirTypeGroup.SUBOBJ_DIR

    @property
    def is_downloadable(self) -> bool:
        return bool(not self.is_dir and self.name)

    def append_to_file(self, fd, path: str):
        fd.write(f"- name: {path}\n")
        fd.write(f"  etag: {self.etag}\n")
        version = self.version
        if version:
            fd.write(f"  version: {self.version}\n")
        fd.write(f"  last_modified: '{time_to_str(self.last_modified)}'\n")
        size = self.size
        fd.write(f"  size: {self.size}\n")
        return size

    def get_metafile_data(self, path: str):
        data: dict[str, Any] = {
            "name": path,
            "etag": self.etag,
        }
        version = self.version
        if version:
            data["version"] = version
        data["last_modified"] = time_to_str(self.last_modified)
        data["size"] = self.size
        return data

    @property
    def full_path(self) -> str:
        if self.is_dir and self.path:
            return self.path + "/"
        return self.path

    def as_uid(self, storage: Optional[StorageURI] = None):
        if storage is None:
            storage = self.source
        return UniqueId(
            storage,
            self.parent,
            self.name,
            self.etag,
            self.size,
            self.vtype,
            self.location,
        )


@attrs.define
class Entry:
    vtype: str = ""
    dir_type: Optional[int] = None
    parent: str = ""
    name: str = ""
    etag: str = ""
    version: str = ""
    is_latest: bool = True
    last_modified: Optional[datetime] = None
    size: int = 0
    owner_name: str = ""
    owner_id: str = ""
    location: Optional[str] = None

    @property
    def is_dir(self) -> bool:
        return self.dir_type == DirType.DIR

    @classmethod
    def from_dir(cls, parent: str, name: str, **kwargs) -> "Entry":
        return cls(dir_type=DirType.DIR, parent=parent, name=name, **kwargs)

    @classmethod
    def from_file(cls, parent: str, name: str, **kwargs) -> "Entry":
        return cls(dir_type=DirType.FILE, parent=parent, name=name, **kwargs)

    @classmethod
    def root(cls):
        return cls(dir_type=DirType.DIR)

    @property
    def path(self) -> str:
        return f"{self.parent}/{self.name}" if self.parent else self.name

    @property
    def full_path(self) -> str:
        if self.is_dir and self.path:
            return self.path + "/"
        return self.path


def get_path(parent: str, name: str):
    return f"{parent}/{name}" if parent else name


@attrs.define
class NodeWithPath:
    n: Node
    path: list[str] = attrs.field(factory=list)

    def append_to_file(self, fd):
        return self.n.append_to_file(fd, "/".join(self.path))

    def get_metafile_data(self):
        return self.n.get_metafile_data("/".join(self.path))

    @property
    def full_path(self) -> str:
        path = "/".join(self.path)
        if self.n.is_dir and path:
            path += "/"
        return path


TIME_FMT = "%Y-%m-%d %H:%M"


def long_line_str(name: str, timestamp: Optional[datetime], owner: str) -> str:
    if timestamp is None:
        time = "-"
    else:
        time = timestamp.strftime(TIME_FMT)
    return f"{owner: <19} {time: <19} {name}"



import asyncio
import functools
import json
import logging
import multiprocessing
import os
import posixpath
import re
from abc import ABC, abstractmethod
from collections.abc import Iterator
from datetime import datetime
from shutil import copy2
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    ClassVar,
    NamedTuple,
    Optional,
)

from botocore.exceptions import ClientError
from dvc_objects.fs.system import reflink
from fsspec.asyn import get_loop, sync
from fsspec.callbacks import DEFAULT_CALLBACK, Callback
from tqdm import tqdm

from dvcx.cache import DVCXCache, UniqueId
from dvcx.client.fileslice import FileSlice, FileWrapper
from dvcx.error import ClientError as DVCXClientError
from dvcx.node import Entry
from dvcx.nodes_fetcher import NodesFetcher
from dvcx.nodes_thread_pool import NodeChunk
from dvcx.storage import StorageURI
from dvcx.utils import TIME_ZERO

if TYPE_CHECKING:
    from fsspec.spec import AbstractFileSystem

    from dvcx.data_storage import AbstractMetastore

logger = logging.getLogger("dvcx")

FETCH_WORKERS = 100
DELIMITER = "/"  # Path delimiter.

DATA_SOURCE_URI_PATTERN = re.compile(r"^[\w]+:\/\/.*$")


class Bucket(NamedTuple):
    name: str
    uri: StorageURI
    created: Optional[datetime]


class Client(ABC):
    MAX_THREADS = multiprocessing.cpu_count()
    FS_CLASS: ClassVar[type["AbstractFileSystem"]]
    PREFIX: ClassVar[str]
    protocol: ClassVar[str]

    def __init__(self, name: str, fs: "AbstractFileSystem", cache: DVCXCache) -> None:
        self.name = name
        self.fs = fs
        self.cache = cache

    @staticmethod
    def get_implementation(url: str) -> type["Client"]:
        from .azure import AzureClient
        from .gcs import GCSClient
        from .local import FileClient
        from .s3 import ClientS3

        if url.lower().startswith(ClientS3.PREFIX):
            return ClientS3
        if url.lower().startswith(GCSClient.PREFIX):
            return GCSClient
        if url.lower().startswith(AzureClient.PREFIX):
            return AzureClient
        return FileClient

    @staticmethod
    def is_data_source_uri(name: str) -> bool:
        # Returns True if name is one of supported data sources URIs, e.g s3 bucket
        return DATA_SOURCE_URI_PATTERN.match(name) is not None

    @staticmethod
    def parse_url(
        source: str,
        metastore: "AbstractMetastore",
        cache: DVCXCache,
        **kwargs,
    ) -> tuple["Client", str]:
        cls = Client.get_implementation(source)
        storage_url, rel_path = cls.split_url(source)
        client = cls.from_name(storage_url, metastore, cache, kwargs)
        return client, rel_path

    @classmethod
    def create_fs(cls, **kwargs) -> "AbstractFileSystem":
        kwargs.setdefault("version_aware", True)
        fs = cls.FS_CLASS(**kwargs)
        fs.invalidate_cache()
        return fs

    @classmethod
    def from_name(
        cls,
        name: str,
        metastore: "AbstractMetastore",
        cache: DVCXCache,
        kwargs: dict[str, Any],
    ) -> "Client":
        return cls(name, cls.create_fs(**kwargs), cache)

    @classmethod
    def from_source(
        cls,
        uri: StorageURI,
        cache: DVCXCache,
        **kwargs,
    ) -> "Client":
        fs = cls.create_fs(**kwargs)
        return cls(fs._strip_protocol(uri), fs, cache)

    @classmethod
    def ls_buckets(cls, **kwargs) -> Iterator[Bucket]:
        for entry in cls.create_fs(**kwargs).ls(cls.PREFIX, detail=True):
            name = entry["name"].rstrip("/")
            yield Bucket(
                name=name,
                uri=StorageURI(f"{cls.PREFIX}{name}"),
                created=entry.get("CreationDate"),
            )

    @classmethod
    def is_root_url(cls, url) -> bool:
        return url == cls.PREFIX

    @property
    def uri(self) -> StorageURI:
        return StorageURI(f"{self.PREFIX}{self.name}")

    @classmethod
    def split_url(cls, url: str) -> tuple[str, str]:
        fill_path = url[len(cls.PREFIX) :]
        path_split = fill_path.split("/", 1)
        bucket = path_split[0]
        path = path_split[1] if len(path_split) > 1 else ""
        return bucket, path

    def url(self, path: str, expires: int = 3600, **kwargs) -> str:
        return self.fs.sign(self.get_full_path(path), expiration=expires, **kwargs)

    async def get_current_etag(self, uid: UniqueId) -> str:
        info = await self.fs._info(self.get_full_path(uid.path))
        return self.convert_info(info, "").etag

    async def get_size(self, path: str) -> int:
        return await self.fs._size(path)

    async def get_file(self, lpath, rpath, callback):
        return await self.fs._get_file(lpath, rpath, callback=callback)

    async def scandir(self, start_prefix):
        result_queue = asyncio.Queue()
        loop = get_loop()
        main_task = loop.create_task(self._fetch(start_prefix, result_queue))
        while (entry := await result_queue.get()) is not None:
            yield entry
        await main_task

    async def _fetch(self, start_prefix="", result_queue=None):
        progress_bar = tqdm(desc=f"Listing {self.uri}", unit=" objects")
        loop = get_loop()

        queue = asyncio.Queue()
        queue.put_nowait(start_prefix)

        async def worker(queue) -> None:
            while True:
                prefix = await queue.get()
                try:
                    subdirs = await self._fetch_dir(prefix, progress_bar, result_queue)
                    for subdir in subdirs:
                        queue.put_nowait(subdir)
                except Exception:
                    while not queue.empty():
                        queue.get_nowait()
                        queue.task_done()
                    raise

                finally:
                    queue.task_done()

        try:
            workers = []
            for _ in range(FETCH_WORKERS):
                workers.append(loop.create_task(worker(queue)))

            # Wait for all fetch tasks to complete
            await queue.join()
            # Stop the workers
            excs = []
            for worker in workers:
                if worker.done() and (exc := worker.exception()):
                    excs.append(exc)
                else:
                    worker.cancel()
            if excs:
                raise excs[0]
        except ClientError as exc:
            raise DVCXClientError(
                exc.response.get("Error", {}).get("Message") or exc,
                exc.response.get("Error", {}).get("Code"),
            ) from exc
        finally:
            # This ensures the progress bar is closed before any exceptions are raised
            progress_bar.close()
            result_queue.put_nowait(None)

    async def _fetch_dir(self, prefix, pbar, result_queue) -> set[str]:
        path = f"{self.name}/{prefix}"
        infos = await self.ls_dir(path)
        files = []
        subdirs = set()
        for info in infos:
            full_path = info["name"]
            subprefix = self.rel_path(full_path)
            if prefix.strip(DELIMITER) == subprefix.strip(DELIMITER):
                continue
            if info["type"] == "directory":
                name = full_path.split(DELIMITER)[-1]
                await result_queue.put(
                    [Entry.from_dir(prefix, name, last_modified=TIME_ZERO)]
                )
                subdirs.add(subprefix)
            else:
                files.append(self.convert_info(info, prefix))
        if files:
            await result_queue.put(files)
        found_count = len(subdirs) + len(files)
        pbar.update(found_count)
        return subdirs

    async def ls_dir(self, path):
        return await self.fs._ls(path, detail=True, versions=True)

    def rel_path(self, path: str) -> str:
        return self.fs.split_path(path)[1]

    def get_full_path(self, rel_path: str) -> str:
        return f"{self.PREFIX}{self.name}/{rel_path}"

    @abstractmethod
    def convert_info(self, v: dict[str, Any], parent: str) -> Entry: ...

    def fetch_nodes(
        self,
        nodes,
        shared_progress_bar=None,
    ) -> None:
        fetcher = NodesFetcher(self, self.MAX_THREADS, self.cache)
        chunk_gen = NodeChunk(self.cache, self.uri, nodes)
        fetcher.run(chunk_gen, shared_progress_bar)

    def instantiate_object(
        self,
        uid: UniqueId,
        dst: str,
        progress_bar: tqdm,
        force: bool = False,
    ) -> None:
        if os.path.exists(dst):
            if force:
                os.remove(dst)
            else:
                progress_bar.close()
                raise FileExistsError(f"Path {dst} already exists")
        self.do_instantiate_object(uid, dst)

    def do_instantiate_object(self, uid: "UniqueId", dst: str) -> None:
        src = self.cache.get_path(uid)
        assert src is not None

        try:
            reflink(src, dst)
        except OSError:
            # Default to copy if reflinks are not supported
            copy2(src, dst)

    def open_object(
        self, uid: UniqueId, use_cache: bool = True, cb: Callback = DEFAULT_CALLBACK
    ) -> BinaryIO:
        """Open a file, including files in tar archives."""
        if uid.vtype == "tar":
            return self._open_tar(uid, use_cache=True)
        if use_cache and (cache_path := self.cache.get_path(uid)):
            return open(cache_path, mode="rb")  # noqa: SIM115
        return FileWrapper(self.fs.open(self.get_full_path(uid.path)), cb)  # type: ignore[return-value]

    def _open_tar(self, uid: UniqueId, use_cache: bool = True):
        assert uid.location is not None
        loc_stack = (
            json.loads(uid.location) if isinstance(uid.location, str) else uid.location
        )
        if len(loc_stack) > 1:
            raise NotImplementedError("Nested v-objects are not supported yet.")
        obj_location = loc_stack[0]
        tar_path = posixpath.split(obj_location["parent"])
        parent_etag = obj_location["etag"]
        offset = obj_location["offset"]
        size = obj_location["size"]
        parent_uid = UniqueId(
            uid.storage,
            *tar_path,
            etag=parent_etag,
            size=-1,
            vtype="",
            location=None,
        )
        f = self.open_object(parent_uid, use_cache=use_cache)
        return FileSlice(f, offset, size, posixpath.basename(uid.path))

    def download(self, uid: UniqueId, *, callback: Callback = DEFAULT_CALLBACK) -> None:
        sync(get_loop(), functools.partial(self._download, uid, callback=callback))

    async def _download(self, uid: UniqueId, *, callback: "Callback" = None) -> None:
        if self.cache.contains(uid):
            # Already in cache, so there's nothing to do.
            return
        await self._put_in_cache(uid, callback=callback)

    def put_in_cache(self, uid: UniqueId, *, callback: "Callback" = None) -> None:
        sync(get_loop(), functools.partial(self._put_in_cache, uid, callback=callback))

    async def _put_in_cache(
        self, uid: UniqueId, *, callback: "Callback" = None
    ) -> None:
        if uid.vtype == "tar":
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(
                None, functools.partial(self._download_from_tar, uid, callback=callback)
            )
            # self._download_from_tar(uid, callback=callback)
            return
        if uid.etag:
            etag = await self.get_current_etag(uid)
            if uid.etag != etag:
                raise FileNotFoundError(
                    f"Invalid etag for {uid.storage}/{uid.path}: "
                    f"expected {uid.etag}, got {etag}"
                )
        await self.cache.download(uid, self, callback=callback)

    def _download_from_tar(self, uid, *, callback: "Callback" = None):
        with self._open_tar(uid, use_cache=False) as f:
            contents = f.read()
        self.cache.store_data(uid, contents)



import os
import shutil
from subprocess import check_output

import pytest
import virtualenv
from dulwich.porcelain import clone
from packaging import version


@pytest.fixture
def bucket():
    return "s3://noaa-bathymetry-pds/"


def pytest_generate_tests(metafunc):
    str_revs = metafunc.config.getoption("--dvcx-revs")
    revs = str_revs.split(",") if str_revs else [None]
    if "dvcx_rev" in metafunc.fixturenames:
        metafunc.parametrize("dvcx_rev", revs, scope="session")


class VirtualEnv:
    def __init__(self, path) -> None:
        self.path = path
        self.bin = self.path / ("Scripts" if os.name == "nt" else "bin")

    def create(self) -> None:
        virtualenv.cli_run([os.fspath(self.path)])

    def run(self, cmd: str, *args: str, env=None) -> None:
        exe = self.which(cmd)
        check_output([exe, *args], env=env)  # noqa: S603

    def which(self, cmd: str) -> str:
        assert self.bin.exists()
        return shutil.which(cmd, path=self.bin) or cmd


@pytest.fixture(scope="session", name="make_dvcx_venv")
def fixture_make_dvcx_venv(tmp_path_factory):
    def _make_dvcx_venv(name):
        venv_dir = tmp_path_factory.mktemp(f"dvcx-venv-{name}")
        venv = VirtualEnv(venv_dir)
        venv.create()
        return venv

    return _make_dvcx_venv


@pytest.fixture(scope="session", name="dvcx_venvs")
def fixture_dvcx_venvs():
    return {}


@pytest.fixture(scope="session", name="dvcx_git_repo")
def fixture_dvcx_git_repo(tmp_path_factory, test_config):
    url = test_config.dvcx_git_repo

    if os.path.isdir(url):
        return url

    tmp_path = os.fspath(tmp_path_factory.mktemp("dvcx-git-repo"))
    clone(url, tmp_path)

    return tmp_path


@pytest.fixture(scope="session", name="dvcx_bin")
def fixture_dvcx_bin(
    dvcx_rev,
    dvcx_venvs,
    make_dvcx_venv,
    dvcx_git_repo,
    test_config,
):
    if dvcx_rev:
        venv = dvcx_venvs.get(dvcx_rev)
        if not venv:
            venv = make_dvcx_venv(dvcx_rev)
            venv.run("pip", "install", "-U", "pip")
            venv.run("pip", "install", f"git+file://{dvcx_git_repo}@{dvcx_rev}")
            dvcx_venvs[dvcx_rev] = venv
        dvcx_bin = venv.which("dvcx")
    else:
        dvcx_bin = test_config.dvcx_bin

    def _dvcx_bin(*args):
        return check_output([dvcx_bin, *args], text=True)  # noqa: S603

    actual = version.parse(_dvcx_bin("--version"))
    _dvcx_bin.version = (actual.major, actual.minor, actual.micro)

    return _dvcx_bin


@pytest.fixture(scope="function", name="make_bench")
def fixture_make_bench(request):
    def _make_bench(name):
        import pytest_benchmark.plugin

        # hack from https://github.com/ionelmc/pytest-benchmark/issues/166
        bench = pytest_benchmark.plugin.benchmark.__pytest_wrapped__.obj(request)

        suffix = f"-{name}"

        def add_suffix(_name):
            start, sep, end = _name.partition("[")
            return start + suffix + sep + end

        bench.name = add_suffix(bench.name)
        bench.fullname = add_suffix(bench.fullname)

        return bench

    return _make_bench


@pytest.fixture(
    scope="function", params=[pytest.param(None, marks=pytest.mark.benchmark)]
)
def bench_dvcx(dvcx_bin, make_bench):
    def _bench_dvcx(*args, **kwargs):
        name = kwargs.pop("name", None)
        name = f"-{name}" if name else ""
        bench = make_bench(args[0] + name)
        return bench.pedantic(dvcx_bin, args=args, **kwargs)

    return _bench_dvcx



import dataclasses
import io
import math
import os
import tarfile
from string import printable
from tarfile import DIRTYPE, TarInfo
from time import sleep, time
from typing import Any, Callable, Optional

import pytest

from dvcx.catalog.catalog import Catalog
from dvcx.dataset import DatasetDependency, DatasetRecord
from dvcx.node import Entry
from dvcx.query import C, DatasetQuery
from dvcx.query.builtins import index_tar
from dvcx.storage import Status as StorageStatus


def make_index(catalog, src: str, entries, ttl: int = 1234):
    lst, _ = catalog.enlist_source(src, ttl, skip_indexing=True)
    lst.insert_entry(Entry.root())
    lst.insert_entries(entries)
    lst.insert_entries_done()
    lst.metastore.mark_storage_indexed(
        src,
        StorageStatus.COMPLETE,
        ttl=ttl,
        prefix="",
        partial_id=lst.metastore.partial_id,
    )


DEFAULT_TREE: dict[str, Any] = {
    "description": "Cats and Dogs",
    "cats": {
        "cat1": "meow",
        "cat2": "mrow",
    },
    "dogs": {
        "dog1": "woof",
        "dog2": "arf",
        "dog3": "bark",
        "others": {"dog4": "ruff"},
    },
}
NUM_TREE = {f"{i:06d}": f"{i}" for i in range(1024)}


def instantiate_tree(path, tree):
    for key, value in tree.items():
        if isinstance(value, str):
            (path / key).write_text(value)
        elif isinstance(value, bytes):
            (path / key).write_bytes(value)
        elif isinstance(value, dict):
            (path / key).mkdir()
            instantiate_tree(path / key, value)
        else:
            raise TypeError(f"{value=}")


def tree_from_path(path, binary=False):
    tree = {}
    for child in path.iterdir():
        if child.is_dir():
            tree[child.name] = tree_from_path(child, binary)
        else:  # noqa: PLR5501
            if binary:
                tree[child.name] = child.read_bytes()
            else:
                tree[child.name] = child.read_text()
    return tree


def uppercase_scheme(uri: str) -> str:
    """
    Makes scheme (or protocol) of an url uppercased
    e.g s3://bucket_name -> S3://bucket_name
    """
    return f'{uri.split(":")[0].upper()}:{":".join(uri.split(":")[1:])}'


def make_tar(tree) -> bytes:
    with io.BytesIO() as tmp:
        with tarfile.open(fileobj=tmp, mode="w") as archive:
            write_tar(tree, archive)
        return tmp.getvalue()


def write_tar(tree, archive, curr_dir=""):
    for key, value in tree.items():
        name = f"{curr_dir}/{key}" if curr_dir else key
        if isinstance(value, str):
            value = value.encode("utf-8")
        if isinstance(value, bytes):
            info = TarInfo(name)
            info.size = len(value)
            f = io.BytesIO(value)
            archive.addfile(info, f)
        elif isinstance(value, dict):
            info = TarInfo(name)
            info.type = DIRTYPE
            archive.addfile(info, io.BytesIO())
            write_tar(value, archive, name)


TARRED_TREE: dict[str, Any] = {"animals.tar": make_tar(DEFAULT_TREE)}


def create_tar_dataset(catalog, uri: str, ds_name: str) -> DatasetQuery:
    """
    Create a dataset from a storage location containing tar archives and other files.

    The resulting dataset contains both the original files (as regular objects)
    and the tar members (as v-objects).
    """
    ds1 = DatasetQuery(path=uri, catalog=catalog)
    tar_entries = ds1.filter(C("name").glob("*.tar")).generate(index_tar)
    return ds1.filter(~C("name").glob("*.tar")).union(tar_entries).save(ds_name)


def skip_if_not_sqlite():
    if os.environ.get("DVCX_METASTORE") or os.environ.get("DVCX_WAREHOUSE"):
        pytest.skip("This test is not supported on other data storages")


WEBFORMAT_TREE: dict[str, Any] = {
    "f1.raw": "raw data",
    "f1.json": '{"similarity": 0.001, "md5": "deadbeef"}',
    "f2.raw": "raw data",
    "f2.json": '{"similarity": 0.005, "md5": "foobar"}',
}


def text_embedding(text: str) -> list[float]:
    """
    Compute a simple text embedding based on character counts.

    These aren't the most meaningful, but will produce a 100-element
    vector of floats between 0 and 1 where texts with similar
    character counts will have similar embeddings. Useful for writing
    unit tests without loading a heavy ML model.
    """
    emb = {c: 0.01 for c in printable}
    for c in text:
        try:
            emb[c] += 1.0
        except KeyError:
            pass
    # sqeeze values between 0 and 1 with an adjusted sigmoid function
    return [2.0 / (1.0 + math.e ** (-x)) - 1.0 for x in emb.values()]


def adjusted_float_diff(a: float, b: float, tol: float = 1e-8):
    """
    Compute how far beyond `tol` the difference between `a` and `b` is.
    Useful for verifying that two floats are close to each other.

    Example:
        >>> adjusted_float_diff(1.2, 1.29, 0.1)
        0.0
        >>> adjusted_float_diff(0.2, 0.8, 0.25)
        0.3500000000000001
    """
    return max(0.0, abs(a - b) - tol)


SIMPLE_DS_QUERY_RECORDS = [
    {
        "parent": "",
        "name": "description",
        "vtype": "",
        "dir_type": 0,
        "is_latest": 1,
        "size": 13,
    },
    {
        "parent": "cats",
        "name": "cat1",
        "vtype": "",
        "dir_type": 0,
        "is_latest": 1,
        "size": 4,
    },
    {
        "parent": "cats",
        "name": "cat2",
        "vtype": "",
        "dir_type": 0,
        "is_latest": 1,
        "size": 4,
    },
    {
        "parent": "dogs",
        "name": "dog1",
        "vtype": "",
        "dir_type": 0,
        "is_latest": 1,
        "size": 4,
    },
    {
        "parent": "dogs",
        "name": "dog2",
        "vtype": "",
        "dir_type": 0,
        "is_latest": 1,
        "size": 3,
    },
    {
        "parent": "dogs",
        "name": "dog3",
        "vtype": "",
        "dir_type": 0,
        "is_latest": 1,
        "size": 4,
    },
    {
        "parent": "dogs/others",
        "name": "dog4",
        "vtype": "",
        "dir_type": 0,
        "is_latest": 1,
        "size": 4,
    },
]


def get_simple_ds_query(path, catalog):
    return (
        DatasetQuery(path=path, catalog=catalog)
        .select(C.parent, C.name, C.vtype, C.dir_type, C.is_latest, C.size)
        .order_by(C.source, C.parent, C.name)
    )


def dataset_dependency_asdict(
    dep: Optional[DatasetDependency],
) -> Optional[dict[str, Any]]:
    """
    Converting to dict with making sure we don't have any additional fields
    that could've been added with subclasses
    """
    if not dep:
        return None
    parsed = {
        k: dataclasses.asdict(dep)[k] for k in DatasetDependency.__dataclass_fields__
    }

    if dep.dependencies:
        parsed["dependencies"] = [
            dataset_dependency_asdict(d) for d in dep.dependencies
        ]

    return parsed


def wait_for_condition(
    callback: Callable,
    message: str,
    check_interval: float = 0.01,
    timeout: float = 1.0,
) -> Any:
    start_time = time()
    while time() - start_time < timeout:
        result = callback()
        if result:
            return result
        sleep(check_interval)
    raise TimeoutError(f"Timeout expired while waiting for: {message}")


def assert_row_names(
    catalog: Catalog, dataset: DatasetRecord, version: int, expected_names: set
) -> None:
    dataset_rows = catalog.ls_dataset_rows(dataset.name, version, limit=20)
    assert dataset_rows
    preview = dataset.get_version(version).preview
    assert preview

    assert (
        {r.name for r in dataset_rows}
        == {r.get("name") for r in preview}
        == expected_names
    )



import base64
import pickle

from dvcx.data_storage.serializer import deserialize
from dvcx.data_storage.sqlite import (
    SQLiteDatabaseEngine,
    SQLiteIDGenerator,
    SQLiteMetastore,
)
from dvcx.storage import StorageURI


def test_sqlite_metastore():
    db = SQLiteDatabaseEngine.from_db_file(":memory:")
    id_generator = SQLiteIDGenerator(db, table_prefix="prefix")
    uri = StorageURI("s3://bucket")

    obj = SQLiteMetastore(id_generator, uri, 1, db)
    assert obj.id_generator == id_generator
    assert obj.uri == uri
    assert obj.partial_id == 1
    assert obj.db == db

    # Test clone
    obj2 = obj.clone()
    assert isinstance(obj2, SQLiteMetastore)
    assert obj2.id_generator.db.db_file == obj.id_generator.db.db_file
    assert obj2.id_generator._table_prefix == obj.id_generator._table_prefix
    assert obj2.uri == uri
    assert obj2.partial_id == 1
    assert obj2.db.db_file == db.db_file
    assert obj2.clone_params() == obj.clone_params()

    # Test serialization
    serialized = obj.serialize()
    assert serialized
    serialized_pickled = base64.b64decode(serialized.encode())
    assert serialized_pickled
    (f, args, kwargs) = pickle.loads(serialized_pickled)  # noqa: S301
    assert f == SQLiteMetastore.init_after_clone
    assert args == []
    assert kwargs == {
        "id_generator_clone_params": id_generator.clone_params(),
        "uri": uri,
        "partial_id": 1,
        "db_clone_params": db.clone_params(),
    }

    # Test deserialization
    obj3 = deserialize(serialized)
    assert isinstance(obj3, SQLiteMetastore)
    assert obj3.id_generator.db.db_file == id_generator.db.db_file
    assert obj3.id_generator._table_prefix == id_generator._table_prefix
    assert obj3.uri == uri
    assert obj3.partial_id == 1
    assert obj3.db.db_file == db.db_file
    assert obj3.clone_params() == obj.clone_params()



from dvcx.sql import select
from dvcx.sql.functions import rand


def test_rand(warehouse):
    query = select(rand())
    result = tuple(warehouse.db.execute(query))
    assert isinstance(result[0][0], int)



"""
UDF to create 'class' and 'type' signals from the directory structure.

To install script dependencies: pip install tabulate
"""

from pathlib import Path

from sqlalchemy import Text
from tabulate import tabulate

from dvcx.query import C, DatasetQuery, udf


@udf("parent", {"class": Text, "type": Text})
def dir_as_class(parent):
    try:
        s_class, s_type = Path(parent).parts[-2:]
    except ValueError:
        return ("", "")
    return (s_class, s_type)


if __name__ == "__main__":
    #   - save as a new dataset
    DatasetQuery(path="gcs://dvcx-zalando-hd-resized/zalando-hd-resized").filter(
        C.name.glob("*.jpg")
    ).add_signals(dir_as_class).save("zalando-with-signals")

    # Output the contents of the new dataset.
    print(tabulate(DatasetQuery(name="zalando-with-signals").results()[:10]))



import pandas as pd

from dvcx.lib.dataset import C, Dataset
from dvcx.lib.file import FileInfo
from dvcx.lib.webdataset import WebDataset
from dvcx.lib.webdataset_laion import WDSLaion
from dvcx.lib.webdataset_meta import LaionMeta, MergeParquetAndNpz
from dvcx.sql.functions import path

wds = (
    Dataset("gcs://dvcx-datacomp-small/shards")
    .filter(C.name.glob("00000000.tar"))
    .generate(WebDataset(spec=WDSLaion), cache=True)
)

meta = (
    Dataset("gcs://dvcx-datacomp-small/metadata")
    .filter(C.name.glob("0020f*"))
    .aggregate(MergeParquetAndNpz(), partition_by=path.file_stem(C.name), cache=True)
    .select_except(FileInfo)
)

res = wds.merge(meta, on=WDSLaion.json.uid, right_on=LaionMeta.uid)

df = res.limit(10).to_pandas()
with pd.option_context("display.max_columns", None):
    print(df)



import concurrent
import concurrent.futures
import threading
from abc import ABC, abstractmethod


class NodeChunk:
    def __init__(
        self, cache, storage, nodes, size_limit=10 * 1024 * 1024, file_limit=100
    ):
        self.cache = cache
        self.storage = storage
        self.nodes = nodes
        self.size_limit = size_limit
        self.file_limit = file_limit

    def __iter__(self):
        return self

    def next_downloadable(self):
        node = next(self.nodes, None)
        while node and (
            not node.is_downloadable or self.cache.contains(node.as_uid(self.storage))
        ):
            node = next(self.nodes, None)
        return node

    def __next__(self):
        node = self.next_downloadable()

        total_size = 0
        total_files = 0
        bucket = []

        while (
            node
            and total_size + node.size < self.size_limit
            and total_files + 1 < self.file_limit
        ):
            bucket.append(node)
            total_size += node.size
            total_files += 1
            node = self.next_downloadable()

        if node:
            bucket.append(node)
            total_size += node.size
            total_files += 1

        if bucket:
            return bucket
        raise StopIteration


class NodesThreadPool(ABC):
    def __init__(self, max_threads):
        self._max_threads = max_threads
        self._thread_counter = 0
        self._thread_lock = threading.Lock()

    def run(
        self,
        chunk_gen,
        progress_bar=None,
    ):
        results = []
        with concurrent.futures.ThreadPoolExecutor(self._max_threads) as th_pool:
            tasks = set()
            self._thread_counter = 0
            for chunk in chunk_gen:
                while len(tasks) >= self._max_threads:
                    done, _ = concurrent.futures.wait(
                        tasks, timeout=1, return_when="FIRST_COMPLETED"
                    )
                    self.done_task(done)

                    tasks = tasks - done
                    self.update_progress_bar(progress_bar)

                tasks.add(th_pool.submit(self.do_task, chunk))
                self.update_progress_bar(progress_bar)

            while tasks:
                done, _ = concurrent.futures.wait(
                    tasks, timeout=1, return_when="FIRST_COMPLETED"
                )
                task_results = self.done_task(done)
                if task_results:
                    results.extend(task_results)

                tasks = tasks - done
                self.update_progress_bar(progress_bar)

            th_pool.shutdown()

        return results

    def update_progress_bar(self, progress_bar):
        if progress_bar is not None:
            with self._thread_lock:
                if self._thread_counter:
                    progress_bar.update(self._thread_counter)
                    self._thread_counter = 0

    def increase_counter(self, value):
        with self._thread_lock:
            self._thread_counter += value

    @abstractmethod
    def do_task(self, chunk):
        pass

    @abstractmethod
    def done_task(self, done):
        pass



import json
from json import JSONDecodeError
from typing import Literal

import anthropic
from anthropic.types import Usage
from anthropic.types.message import Message
from pydantic import Field

from dvcx.lib.feature import Feature
from dvcx.lib.feature_types import FeatureLike, pydantic_to_feature
from dvcx.lib.feature_udf import FeatureAggregator, FeatureMapper
from dvcx.lib.file import FileInfo, TextFile

default_model_name = "claude-3-haiku-20240307"
DEFAULT_OUTPUT_TOKENS = 1024

MessageFeature = pydantic_to_feature(Message)
UsageFeature = pydantic_to_feature(Usage)


class Claude(MessageFeature):  # type: ignore[valid-type,misc]
    error: str = ""


class Rating(Feature):
    status: Literal["Success", "Failure"] = "Failure"
    explanation: str = ""
    usage: UsageFeature = UsageFeature(input_tokens=0, output_tokens=0)  # type: ignore[valid-type,call-arg]
    error: str = ""


class Summary(Feature):
    reasons: list[str] = Field(default_factory=list)
    usage: UsageFeature = UsageFeature(input_tokens=0, output_tokens=0)  # type: ignore[valid-type,call-arg]
    error: str = ""


class ClaudeClient:
    def __init__(
        self,
        prompt,
        messages=None,
        model_name="",
        api_key=None,
        max_retries=5,
        **kwargs,
    ):
        self.prompt = prompt
        self.model_name = model_name
        self.messages = messages
        self.api_key = api_key
        self.max_retries = max_retries
        self.kwargs = kwargs
        self.model = None

        if not self.messages:
            self.messages = []
        if "max_tokens" not in self.kwargs:
            self.kwargs["max_tokens"] = DEFAULT_OUTPUT_TOKENS

    def bootstrap(self):
        self.model = anthropic.Anthropic(api_key=self.api_key)
        if self.max_retries:
            self.model = self.model.with_options(max_retries=self.max_retries)

    def teardown(self):
        self.model.close()
        self.model = None

    def llm_call(self, text):
        try:
            response = self.model.messages.create(  # type: ignore[attr-defined]
                model=self.model_name,
                system=self.prompt,
                messages=[{"role": "user", "content": text}, *self.messages],
                **self.kwargs,
            )

            return Claude(**response.model_dump())
        except Exception as exc:  # noqa: BLE001
            return Claude(
                id="",
                content=[],
                model="",
                role="assistant",
                type="message",
                usage={"input_tokens": 0, "output_tokens": 0},
                error=str(exc),
            )

    @staticmethod
    def get_rating(claude: Claude):
        if claude.error:
            response = Rating(status="Failure", usage=claude.usage, error=claude.error)
        elif len(claude.content) > 0:
            text = claude.content[0].text
            try:
                j = json.loads(text)
                response = Rating(**j)
                response.usage = claude.usage
            except JSONDecodeError as exc:
                response = Rating(
                    error=f"json parsing error '{text}': {exc!s}", usage=claude.usage
                )
        else:
            response = Rating(
                status="Failure", error="no responses", usage=claude.usage
            )
        return response


class ClaudeMessage(ClaudeClient, FeatureMapper):
    def __init__(
        self,
        data: FeatureLike = TextFile,
        **kwargs,
    ):
        ClaudeClient.__init__(self, **kwargs)
        FeatureMapper.__init__(self, data, Claude)

    def process(self, data):
        return self.llm_call(data.get_value())


class TextAnalytics(ClaudeClient, FeatureMapper):
    """Run system prompt on records one-by-one and return responses in a column"""

    def __init__(self, data: FeatureLike = TextFile, **kwargs):
        ClaudeClient.__init__(self, **kwargs)
        FeatureMapper.__init__(self, data, Rating)

    def process(self, data):
        claude = self.llm_call(data.get_value())
        return self.get_rating(claude)


class AggregateTextAnalytics(ClaudeClient, FeatureAggregator):
    """Run the system prompt on all records and return one response"""

    def __init__(self, data: FeatureLike = TextFile, **kwargs):
        ClaudeClient.__init__(self, **kwargs)
        FeatureAggregator.__init__(self, data, [FileInfo, Summary])

    def process(self, args):
        all_texts = "\n=====\n".join(data.get_value() for data in args)
        claude = self.llm_call(all_texts)

        if claude.error == "":
            text = claude.content[0].text
            try:
                j = json.loads(text)
                if isinstance(j, list):
                    result = Summary(reasons=j, usage=claude.usage)
                else:
                    result = Summary(
                        error=f"result is not list but '{j}'", usage=claude.usage
                    )
            except JSONDecodeError as exc:
                result = Summary(
                    error=f"json parsing error '{text}': {exc!s}", usage=claude.usage
                )
        else:
            result = Summary(error=claude.error, usage=claude.usage)

        yield FileInfo(name="", vtype="claude-agg"), result



import inspect
from collections.abc import Iterable, Iterator, Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Optional,
    TypeVar,
)

import sqlalchemy as sa
from sqlalchemy.sql import func as f
from sqlalchemy.sql.expression import null, true

from dvcx.node import DirType, DirTypeGroup
from dvcx.sql.functions import path
from dvcx.sql.types import JSON, Boolean, DateTime, Int, Int64, SQLType, String

if TYPE_CHECKING:
    from sqlalchemy import Engine
    from sqlalchemy.engine.interfaces import Dialect
    from sqlalchemy.sql.base import Executable, ReadOnlyColumnCollection
    from sqlalchemy.sql.elements import KeyedColumnElement
    from sqlalchemy.sql.selectable import Select


def dedup_columns(columns: Iterable[sa.Column]) -> list[sa.Column]:
    """
    Removes duplicate columns from a list of columns.
    If column with the same name and different type is found, exception is
    raised
    """
    c_set: dict[str, sa.Column] = {}
    for c in columns:
        if ec := c_set.get(c.name, None):
            if str(ec.type) != str(c.type):
                raise ValueError(
                    f"conflicting types for column {c.name}:"
                    f"{c.type!s} and {ec.type!s}"
                )
            continue
        c_set[c.name] = c

    return list(c_set.values())


def convert_rows_custom_column_types(
    columns: "ReadOnlyColumnCollection[str, KeyedColumnElement[Any]]",
    rows: Iterator[tuple[Any, ...]],
    dialect: "Dialect",
):
    """
    This function converts values of rows columns based on their types which are
    defined in columns. We are only converting column values for which types are
    subclasses of our SQLType, as only for those we have converters registered.
    """
    # indexes of SQLType column in a list of columns so that we can skip the rest
    custom_columns_types: list[tuple[int, SQLType]] = [
        (idx, c.type) for idx, c in enumerate(columns) if isinstance(c.type, SQLType)
    ]

    if not custom_columns_types:
        yield from rows

    for row in rows:
        row_list = list(row)
        for idx, t in custom_columns_types:
            row_list[idx] = t.on_read_convert(row_list[idx], dialect)

        yield tuple(row_list)


class Table:
    def __init__(self, name: str, metadata: Optional["sa.MetaData"] = None):
        self.metadata: sa.MetaData = metadata if metadata is not None else sa.MetaData()
        self.name: str = name

    def adjust_default_column_types(self, table: "sa.Table") -> None:
        """
        Adjusting types of default columns to be instances of our SQLType since
        when getting table by reflection, that information is lost
        """
        default_columns = {c.name: c for c in self.default_columns}
        for c in table.c:
            c.type = default_columns.get(c.name, c).type

    @property
    def columns(self) -> "ReadOnlyColumnCollection[str, sa.Column[Any]]":
        return self.table.columns

    @property
    def c(self):
        return self.columns

    @classmethod
    def dataset_default_columns(cls) -> list[sa.Column]:
        return []

    @property
    def custom_columns(self) -> list[sa.Column]:
        return [c for c in self.table.c if c.name not in DATASET_CORE_COLUMN_NAMES]

    @property
    def default_columns(self) -> list[sa.Column]:
        return [c for c in self.table.c if c.name in DATASET_CORE_COLUMN_NAMES]

    @property
    def table(self) -> "sa.Table":
        return self.get_table()

    def get_table(self) -> "sa.Table":
        table = self.metadata.tables.get(self.name)
        if table is None:
            table = sa.Table(
                self.name,
                self.metadata,
                *self.default_columns,
            )
        self.adjust_default_column_types(table)

        return table

    def apply_conditions(self, query: "Executable") -> "Executable":
        """
        Apply any conditions that belong on all selecting queries.

        This could be used to filter tables that use access control.
        """
        return query

    def select(self, *columns):
        if not columns:
            query = self.table.select()
        else:
            query = sa.select(*columns).select_from(self.table)
        return self.apply_conditions(query)

    def insert(self):
        return self.table.insert()

    def update(self):
        return self.apply_conditions(self.table.update())

    def delete(self):
        return self.apply_conditions(self.table.delete())


class DirExpansion:
    @staticmethod
    def base_select(q):
        return sa.select(
            q.c.id,
            q.c.vtype,
            (q.c.dir_type == DirType.DIR).label("is_dir"),
            q.c.source,
            q.c.parent,
            q.c.name,
            q.c.version,
            q.c.location,
        )

    @staticmethod
    def apply_group_by(q):
        return (
            sa.select(
                f.min(q.c.id).label("id"),
                q.c.vtype,
                q.c.is_dir,
                q.c.source,
                q.c.parent,
                q.c.name,
                q.c.version,
                f.max(q.c.location).label("location"),
            )
            .select_from(q)
            .group_by(
                q.c.source, q.c.parent, q.c.name, q.c.vtype, q.c.is_dir, q.c.version
            )
            .order_by(
                q.c.source, q.c.parent, q.c.name, q.c.vtype, q.c.is_dir, q.c.version
            )
        )

    @classmethod
    def query(cls, q):
        q = cls.base_select(q).cte(recursive=True)
        parent_parent = path.parent(q.c.parent)
        parent_name = path.name(q.c.parent)
        q = q.union_all(
            sa.select(
                sa.literal(-1).label("id"),
                sa.literal("").label("vtype"),
                true().label("is_dir"),
                q.c.source,
                parent_parent.label("parent"),
                parent_name.label("name"),
                sa.literal("").label("version"),
                null().label("location"),
            ).where((parent_name != "") | (parent_parent != ""))
        )
        return cls.apply_group_by(q)


class DatasetRow(Table):
    dataset_dir_expansion = DirExpansion.query

    def __init__(
        self,
        name: str,
        engine: "Engine",
        metadata: Optional["sa.MetaData"] = None,
        custom_column_types: Optional[dict[str, SQLType]] = None,
    ):
        self.engine = engine
        self.custom_column_types = custom_column_types
        super().__init__(name, metadata)

    @classmethod
    def dataset_default_columns(cls) -> list[sa.Column]:
        return [
            sa.Column("id", Int, primary_key=True),
            sa.Column("vtype", String, nullable=False, index=True),
            sa.Column("dir_type", Int, index=True),
            sa.Column("parent", String, index=True),
            sa.Column("name", String, nullable=False, index=True),
            sa.Column("etag", String),
            sa.Column("version", String),
            sa.Column("is_latest", Boolean),
            sa.Column("last_modified", DateTime(timezone=True)),
            sa.Column("size", Int64, nullable=False, index=True),
            sa.Column("owner_name", String),
            sa.Column("owner_id", String),
            sa.Column("random", Int64, nullable=False),
            sa.Column("location", JSON),
            sa.Column("source", String, nullable=False),
        ]

    @property
    def custom_columns(self) -> list[sa.Column]:
        return [c for c in self.table.c if c.name not in DATASET_CORE_COLUMN_NAMES]

    @property
    def default_columns(self) -> list[sa.Column]:
        return [c for c in self.table.c if c.name in DATASET_CORE_COLUMN_NAMES]

    @staticmethod
    def copy_signal_column(column: sa.Column):
        """
        Copy a sqlalchemy Column object intended for use as a signal column.

        This does not copy all attributes as certain attributes such as
        table are too context-dependent and the purpose of this function is
        adding a signal column from one table to another table.

        We can't use Column.copy() as it only works in certain contexts.
        See https://github.com/sqlalchemy/sqlalchemy/issues/5953
        """
        return sa.Column(
            column.name,
            column.type,
            primary_key=column.primary_key,
            index=column.index,
            nullable=column.nullable,
            default=column.default,
            server_default=column.server_default,
            unique=column.unique,
        )

    @classmethod
    def new_table(
        cls,
        name: str,
        custom_columns: Sequence["sa.Column"] = (),
        metadata: Optional["sa.MetaData"] = None,
    ):
        # copy columns, since re-using the same objects from another table
        # may raise an error
        custom_columns = [cls.copy_signal_column(c) for c in custom_columns]
        if metadata is None:
            metadata = sa.MetaData()
        return sa.Table(
            name,
            metadata,
            *cls.dataset_default_columns(),
            *custom_columns,
        )

    def get_table(self) -> "sa.Table":
        table = self.metadata.tables.get(self.name)
        if table is None:
            table = sa.Table(
                self.name,
                self.metadata,
                extend_existing=True,
                autoload_with=self.engine,
            )

        # adjusting types for custom columns to be instances of SQLType if possible
        if self.custom_column_types:
            for c in table.columns:
                if c.name in self.custom_column_types:
                    t = self.custom_column_types[c.name]
                    c.type = t() if inspect.isclass(t) else t

        return table

    def dir_expansion(self):
        return self.dataset_dir_expansion(self)

    def dataset_query(
        self,
        *column_names: str,
    ) -> "Select":
        if not column_names:
            column_names = DATASET_CORE_COLUMN_NAMES
        column_objects = [self.c[c] for c in column_names]
        # include all object types - file, tar archive, tar file (subobject)
        return self.select(*column_objects).where(
            self.c.dir_type.in_(DirTypeGroup.FILE) & (self.c.is_latest == true())
        )


DATASET_CORE_COLUMN_NAMES = tuple(
    col.name for col in DatasetRow.dataset_default_columns()
)

PARTITION_COLUMN_ID = "partition_id"

partition_col_names = [PARTITION_COLUMN_ID]


def partition_columns() -> Sequence["sa.Column"]:
    return [
        sa.Column(PARTITION_COLUMN_ID, sa.Integer),
    ]


DatasetRowT = TypeVar("DatasetRowT", bound=DatasetRow)


class Schema(Generic[DatasetRowT]):
    dataset_row_cls: type[DatasetRowT]


class DefaultSchema(Schema[DatasetRow]):
    def __init__(self):
        self.dataset_row_cls = DatasetRow



import json
import uuid
from json import dumps
from unittest.mock import ANY

import pytest
import sqlalchemy as sa
from dateutil.parser import isoparse

from dvcx.catalog.catalog import DATASET_INTERNAL_ERROR_MESSAGE
from dvcx.dataset import DatasetDependencyType
from dvcx.dataset import Status as DatasetStatus
from dvcx.error import DatasetInvalidVersionError, DatasetNotFoundError
from dvcx.query import DatasetQuery, udf
from dvcx.sql.types import (
    JSON,
    Array,
    Binary,
    Boolean,
    Float,
    Float32,
    Float64,
    Int,
    Int32,
    Int64,
    SQLType,
    String,
)
from tests.utils import assert_row_names, dataset_dependency_asdict


def add_column(engine, table_name, column, catalog):
    # Simple method that adds new column to a table, with default value if specified
    column_name = column.compile(dialect=engine.dialect)
    column_type = column.type.compile(engine.dialect)
    query_str = f"ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}"
    if column.default:
        query_str += f" DEFAULT {column.default.arg}"
    catalog.warehouse.db.execute_str(query_str)


@pytest.fixture
def empty_registered_dataset(listed_bucket, cloud_test_catalog):
    name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog
    dataset = catalog.create_dataset(name, create_rows=False)

    catalog.metastore.update_dataset_status(
        dataset, DatasetStatus.COMPLETE, version=dataset.current_version
    )

    return catalog.get_dataset(name)


@pytest.mark.parametrize("create_rows", [True, False])
def test_create_dataset_no_version_specified(cloud_test_catalog, create_rows):
    catalog = cloud_test_catalog.catalog

    name = uuid.uuid4().hex
    dataset = catalog.create_dataset(
        name,
        query_script="script",
        custom_columns=[sa.Column("similarity", Float32)],
        create_rows=create_rows,
    )

    assert dataset.versions_values == [1]

    dataset_version = dataset.get_version(1)

    assert dataset.name == name
    assert dataset.query_script == "script"
    assert dataset_version.query_script == "script"
    assert dataset.custom_column_types["similarity"] == Float32
    assert dataset_version.custom_column_types["similarity"] == Float32
    assert dataset_version.status == DatasetStatus.PENDING
    assert dataset.status == DatasetStatus.CREATED  # dataset status is deprecated
    if create_rows:
        assert dataset_version.num_objects == 0
    else:
        assert dataset_version.num_objects is None


@pytest.mark.parametrize("create_rows", [True, False])
def test_create_dataset_with_explicit_version(cloud_test_catalog, create_rows):
    catalog = cloud_test_catalog.catalog

    name = uuid.uuid4().hex
    dataset = catalog.create_dataset(
        name,
        version=1,
        query_script="script",
        custom_columns=[sa.Column("similarity", Float32)],
        create_rows=create_rows,
    )

    assert dataset.versions_values == [1]

    dataset_version = dataset.get_version(1)

    assert dataset.name == name
    assert dataset.query_script == "script"
    assert dataset_version.query_script == "script"
    assert dataset.custom_column_types["similarity"] == Float32
    assert dataset_version.custom_column_types["similarity"] == Float32
    assert dataset_version.status == DatasetStatus.PENDING
    assert dataset.status == DatasetStatus.CREATED
    if create_rows:
        assert dataset_version.num_objects == 0
    else:
        assert dataset_version.num_objects is None


@pytest.mark.parametrize("create_rows", [True, False])
def test_create_dataset_already_exist(cloud_test_catalog, dogs_dataset, create_rows):
    catalog = cloud_test_catalog.catalog

    dataset = catalog.create_dataset(
        dogs_dataset.name,
        query_script="script",
        custom_columns=[sa.Column("similarity", Float32)],
        create_rows=create_rows,
    )

    assert dataset.versions_values == [1, 2]

    dataset_version = dataset.get_version(2)

    assert dataset.name == dogs_dataset.name
    assert dataset_version.query_script == "script"
    assert dataset_version.custom_column_types["similarity"] == Float32
    assert dataset_version.status == DatasetStatus.PENDING
    assert dataset.status == DatasetStatus.COMPLETE
    if create_rows:
        assert dataset_version.num_objects == 0
    else:
        assert dataset_version.num_objects is None


@pytest.mark.parametrize("create_rows", [True, False])
def test_create_dataset_already_exist_wrong_version(
    cloud_test_catalog, dogs_dataset, create_rows
):
    catalog = cloud_test_catalog.catalog

    with pytest.raises(DatasetInvalidVersionError) as exc_info:
        catalog.create_dataset(
            dogs_dataset.name,
            version=1,
            create_rows=create_rows,
        )
    assert str(exc_info.value) == (
        f"Version 1 already exists in dataset {dogs_dataset.name}"
    )


def test_get_dataset(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog

    dataset = catalog.get_dataset(dogs_dataset.name)
    assert dataset.name == dogs_dataset.name

    with pytest.raises(DatasetNotFoundError):
        catalog.get_dataset("wrong name")


# Returns None if the table does not exist
def get_table_row_count(db, table_name):
    if not db.has_table(table_name):
        return None
    query = sa.select(sa.func.count()).select_from(sa.table(table_name))
    return next(db.execute(query), (None,))[0]


def test_create_dataset_from_sources(listed_bucket, cloud_test_catalog):
    dataset_name = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    dataset = catalog.create_dataset_from_sources(
        dataset_name, [f"{src_uri}/dogs/*"], recursive=True
    )

    dataset_version = dataset.get_version(dataset.latest_version)

    assert dataset.name == dataset_name
    assert dataset.description is None
    assert dataset.versions_values == [1]
    assert dataset.labels == []
    assert dataset.status == DatasetStatus.COMPLETE

    assert dataset_version.status == DatasetStatus.COMPLETE
    assert dataset_version.created_at
    assert dataset_version.finished_at
    assert dataset_version.error_message == ""
    assert dataset_version.error_stack == ""
    assert dataset_version.script_output == ""
    assert dataset_version.sources == f"{src_uri}/dogs/*"

    default_dataset_custom_column_types = {
        c.name: c.type.__class__
        for c in catalog.warehouse.dataset_row_cls.dataset_default_columns()
        if isinstance(c.type, SQLType)
    }

    assert dataset.custom_column_types == default_dataset_custom_column_types
    assert dataset.query_script == ""

    assert dataset_version.custom_column_types == default_dataset_custom_column_types
    assert dataset_version.query_script == ""
    assert dataset_version.num_objects
    assert dataset_version.preview


def test_create_dataset_from_sources_empty_sources(cloud_test_catalog):
    dataset_name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog

    with pytest.raises(ValueError) as exc_info:
        catalog.create_dataset_from_sources(dataset_name, [], recursive=True)

    assert str(exc_info.value) == "Sources needs to be non empty list"


def test_create_dataset_from_sources_failed(listed_bucket, cloud_test_catalog, mocker):
    dataset_name = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog
    with mocker.patch.object(
        catalog.warehouse.__class__,
        "create_dataset_rows_table",
        side_effect=RuntimeError("Error"),
    ) as _:
        with pytest.raises(RuntimeError):
            catalog.create_dataset_from_sources(
                dataset_name, [f"{src_uri}/dogs/*"], recursive=True
            )

    dataset = catalog.get_dataset(dataset_name)
    dataset_version = dataset.get_version(dataset.latest_version)

    assert dataset.name == dataset_name
    assert dataset.status == DatasetStatus.FAILED
    assert dataset.versions_values == [1]
    assert dataset.created_at
    assert dataset.finished_at
    assert dataset.error_message == DATASET_INTERNAL_ERROR_MESSAGE
    assert dataset.error_stack
    assert dataset.query_script == ""

    assert dataset_version.status == DatasetStatus.FAILED
    assert dataset_version.created_at
    assert dataset_version.finished_at
    assert dataset_version.error_message == DATASET_INTERNAL_ERROR_MESSAGE
    assert dataset_version.error_stack
    assert dataset_version.sources == f"{src_uri}/dogs/*"
    assert dataset_version.num_objects is None
    assert dataset_version.size is None
    assert dataset_version.preview is None


def test_create_dataset_whole_bucket(listed_bucket, cloud_test_catalog):
    dataset_name_1 = uuid.uuid4().hex
    dataset_name_2 = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    ds1 = catalog.create_dataset_from_sources(
        dataset_name_1, [f"{src_uri}"], recursive=True
    )
    ds2 = catalog.create_dataset_from_sources(
        dataset_name_2, [f"{src_uri}/"], recursive=True
    )

    expected_rows = {
        "description",
        "cat1",
        "cat2",
        "dog1",
        "dog2",
        "dog3",
        "dog4",
    }

    assert_row_names(catalog, ds1, ds1.latest_version, expected_rows)

    assert_row_names(catalog, ds2, ds2.latest_version, expected_rows)


@pytest.mark.parametrize("target_version", [None, 3])
def test_registering_dataset(
    cloud_test_catalog, dogs_dataset, cats_dataset, target_version
):
    catalog = cloud_test_catalog.catalog

    # make sure there is a custom columns inside, other than default ones
    catalog.metastore.update_dataset_version(
        dogs_dataset,
        1,
        sources="s3://ldb-public",
        query_script="DatasetQuery()",
        error_message="no error",
        error_stack="no error stack",
        script_output="log",
    )

    dogs_version = dogs_dataset.get_version(1)

    dataset = catalog.register_dataset(
        dogs_dataset,
        1,
        cats_dataset,
        target_version=target_version,
    )

    # if not provided, it will end up being next dataset version
    target_version = target_version or cats_dataset.next_version

    assert dataset.name == cats_dataset.name
    assert dataset.status == DatasetStatus.COMPLETE
    assert dataset.versions_values == [1, target_version]

    version1 = dataset.get_version(1)
    assert version1.status == DatasetStatus.COMPLETE

    version2 = dataset.get_version(target_version)
    assert version2.status == DatasetStatus.COMPLETE
    assert version2.sources == "s3://ldb-public"
    assert version2.query_script == "DatasetQuery()"
    assert version2.error_message == "no error"
    assert version2.error_stack == "no error stack"
    assert version2.script_output == "log"
    assert version2.custom_column_types == dogs_version.custom_column_types
    assert version2.created_at == dogs_version.created_at
    assert version2.finished_at == dogs_version.finished_at
    assert dogs_version.num_objects
    assert version2.num_objects == dogs_version.num_objects
    assert dogs_version.size
    assert version2.size == dogs_version.size

    assert_row_names(
        catalog,
        dataset,
        1,
        {
            "cat1",
            "cat2",
        },
    )

    assert_row_names(
        catalog,
        dataset,
        target_version,
        {
            "dog1",
            "dog2",
            "dog3",
            "dog4",
        },
    )

    with pytest.raises(DatasetNotFoundError):
        # since it had only one version, it should be completely removed
        catalog.get_dataset(dogs_dataset.name)


@pytest.mark.parametrize("target_version", [None, 3])
def test_registering_dataset_source_dataset_with_multiple_versions(
    cloud_test_catalog, dogs_dataset, cats_dataset, target_version
):
    catalog = cloud_test_catalog.catalog

    # creating one more version for dogs, not to end up completely removed
    dogs_dataset = catalog.create_new_dataset_version(dogs_dataset, 2)

    dataset = catalog.register_dataset(
        dogs_dataset,
        1,
        cats_dataset,
        target_version=target_version,
    )

    # if not provided, it will end up being next dataset version
    target_version = target_version or cats_dataset.next_version

    assert dataset.name == cats_dataset.name
    assert dataset.versions_values == [1, target_version]

    assert_row_names(
        catalog,
        dataset,
        1,
        {
            "cat1",
            "cat2",
        },
    )

    assert_row_names(
        catalog,
        dataset,
        target_version,
        {
            "dog1",
            "dog2",
            "dog3",
            "dog4",
        },
    )

    # check if dogs dataset is still present as it has one more version
    dogs_dataset = catalog.get_dataset(dogs_dataset.name)
    assert dogs_dataset.versions_values == [2]
    dataset_version = dogs_dataset.get_version(2)
    assert dataset_version.num_objects == 0


@pytest.mark.parametrize("target_version", [None, 3])
def test_registering_dataset_with_new_version_of_itself(
    cloud_test_catalog, cats_dataset, target_version
):
    catalog = cloud_test_catalog.catalog

    dataset = catalog.register_dataset(
        cats_dataset,
        1,
        cats_dataset,
        target_version=target_version,
    )

    # if not provided, it will end up being next dataset version
    target_version = target_version or cats_dataset.next_version

    assert dataset.name == cats_dataset.name
    assert dataset.versions_values == [target_version]

    assert_row_names(catalog, dataset, target_version, {"cat1", "cat2"})


def test_registering_dataset_invalid_target_version(
    cloud_test_catalog, cats_dataset, dogs_dataset
):
    catalog = cloud_test_catalog.catalog

    with pytest.raises(DatasetInvalidVersionError) as exc_info:
        catalog.register_dataset(
            dogs_dataset,
            1,
            cats_dataset,
            target_version=1,
        )
    assert str(exc_info.value) == (
        "Version 1 must be higher than the current latest one"
    )


def test_registering_dataset_invalid_source_version(
    cloud_test_catalog, cats_dataset, dogs_dataset
):
    catalog = cloud_test_catalog.catalog

    with pytest.raises(ValueError) as exc_info:
        catalog.register_dataset(
            dogs_dataset,
            5,
            cats_dataset,
            target_version=2,
        )
    assert str(exc_info.value) == (
        f"Dataset {dogs_dataset.name} does not have version 5"
    )


def test_registering_dataset_source_version_in_non_final_status(
    cloud_test_catalog, cats_dataset, dogs_dataset
):
    catalog = cloud_test_catalog.catalog
    catalog.metastore.update_dataset_version(
        dogs_dataset,
        1,
        status=DatasetStatus.PENDING,
    )

    with pytest.raises(ValueError) as exc_info:
        catalog.register_dataset(
            dogs_dataset,
            1,
            cats_dataset,
            target_version=2,
        )
    assert str(exc_info.value) == (
        "Cannot register dataset version in non final status"
    )


def test_remove_dataset(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog

    dataset_version = dogs_dataset.get_version(1)
    assert dataset_version.num_objects

    catalog.remove_dataset(dogs_dataset.name, force=True)
    with pytest.raises(DatasetNotFoundError):
        catalog.get_dataset(dogs_dataset.name)

    dataset_table_name = catalog.warehouse.dataset_table_name(dogs_dataset.name, 1)
    assert get_table_row_count(catalog.warehouse.db, dataset_table_name) is None

    assert catalog.metastore.get_direct_dataset_dependencies(dogs_dataset, 1) == []


def test_remove_dataset_with_multiple_versions(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog

    updated_dogs_dataset = catalog.create_new_dataset_version(dogs_dataset, 2)
    assert updated_dogs_dataset.has_version(2)
    assert updated_dogs_dataset.has_version(1)

    catalog.remove_dataset(updated_dogs_dataset.name, force=True)
    with pytest.raises(DatasetNotFoundError):
        catalog.get_dataset(updated_dogs_dataset.name)

    assert (
        catalog.metastore.get_direct_dataset_dependencies(updated_dogs_dataset, 1) == []
    )


def test_remove_dataset_dataset_not_found(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog

    with pytest.raises(DatasetNotFoundError):
        catalog.remove_dataset("wrong_name", force=True)


def test_remove_dataset_wrong_version(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog

    with pytest.raises(DatasetInvalidVersionError):
        catalog.remove_dataset(dogs_dataset.name, version=100)


def test_edit_dataset(cloud_test_catalog, dogs_dataset):
    dataset_old_name = dogs_dataset.name
    dataset_new_name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog

    catalog.edit_dataset(
        dogs_dataset.name,
        new_name=dataset_new_name,
        description="new description",
        labels=["cats", "birds"],
    )

    dataset = catalog.get_dataset(dataset_new_name)
    assert dataset.versions_values == [1]
    assert dataset.name == dataset_new_name
    assert dataset.description == "new description"
    assert dataset.labels == ["cats", "birds"]

    # check if dataset tables are renamed correctly
    old_dataset_table_name = catalog.warehouse.dataset_table_name(dataset_old_name, 1)
    new_dataset_table_name = catalog.warehouse.dataset_table_name(dataset_new_name, 1)
    assert get_table_row_count(catalog.warehouse.db, old_dataset_table_name) is None
    expected_table_row_count = get_table_row_count(
        catalog.warehouse.db, new_dataset_table_name
    )
    assert expected_table_row_count
    assert dataset.get_version(1).num_objects == expected_table_row_count


def test_edit_dataset_same_name(cloud_test_catalog, dogs_dataset):
    dataset_old_name = dogs_dataset.name
    dataset_new_name = dogs_dataset.name
    catalog = cloud_test_catalog.catalog

    catalog.edit_dataset(dogs_dataset.name, new_name=dataset_new_name)

    dataset = catalog.get_dataset(dataset_new_name)
    assert dataset.name == dataset_new_name

    # check if dataset tables are renamed correctly
    old_dataset_table_name = catalog.warehouse.dataset_table_name(dataset_old_name, 1)
    new_dataset_table_name = catalog.warehouse.dataset_table_name(dataset_new_name, 1)
    expected_table_row_count = get_table_row_count(
        catalog.warehouse.db, old_dataset_table_name
    )
    assert expected_table_row_count
    assert dataset.get_version(1).num_objects == expected_table_row_count
    assert expected_table_row_count == get_table_row_count(
        catalog.warehouse.db, new_dataset_table_name
    )


def test_edit_dataset_remove_labels_and_description(cloud_test_catalog, dogs_dataset):
    dataset_new_name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog

    catalog.edit_dataset(
        dogs_dataset.name,
        new_name=dataset_new_name,
        description="",
        labels=[],
    )

    dataset = catalog.get_dataset(dataset_new_name)
    assert dataset.versions_values == [1]
    assert dataset.name == dataset_new_name
    assert dataset.description == ""
    assert dataset.labels == []


def test_ls_dataset_rows(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog

    assert {r.name for r in catalog.ls_dataset_rows(dogs_dataset.name, 1)} == {
        "dog1",
        "dog2",
        "dog3",
        "dog4",
    }


def test_ls_dataset_rows_with_limit_offset(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog

    # these should be sorted by id already
    all_rows = list(
        catalog.ls_dataset_rows(
            dogs_dataset.name,
            1,
        )
    )

    assert {
        r.name
        for r in catalog.ls_dataset_rows(
            dogs_dataset.name,
            1,
            offset=2,
            limit=1,
        )
    } == {
        all_rows[2].name,
    }


def test_ls_dataset_rows_with_custom_columns(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog
    int_example = 25

    @udf(
        (),
        {
            "int_col": Int,
            "int_col_32": Int32,
            "int_col_64": Int64,
            "float_col": Float,
            "float_col_32": Float32,
            "float_col_64": Float64,
            "array_col": Array(Float),
            "array_col_nested": Array(Array(Float)),
            "array_col_32": Array(Float32),
            "array_col_64": Array(Float64),
            "string_col": String,
            "bool_col": Boolean,
            "json_col": JSON,
            "binary_col": Binary,
        },
    )
    def test_types():
        return (
            5,
            5,
            5,
            0.5,
            0.5,
            0.5,
            [0.5],
            [[0.5], [0.5]],
            [0.5],
            [0.5],
            "s",
            True,
            dumps({"a": 1}),
            int_example.to_bytes(2, "big"),
        )

    (
        DatasetQuery(name=dogs_dataset.name, catalog=catalog)
        .add_signals(test_types)
        .save("dogs_custom_columns")
    )

    for r in catalog.ls_dataset_rows("dogs_custom_columns", 1, custom_columns=True):
        assert r.custom == {
            "int_col": 5,
            "int_col_32": 5,
            "int_col_64": 5,
            "float_col": 0.5,
            "float_col_32": 0.5,
            "float_col_64": 0.5,
            "array_col": [0.5],
            "array_col_nested": [[0.5], [0.5]],
            "array_col_32": [0.5],
            "array_col_64": [0.5],
            "string_col": "s",
            "bool_col": True,
            "json_col": dumps({"a": 1}),
            "binary_col": int_example.to_bytes(2, "big"),
        }


def test_dataset_preview_custom_columns(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog
    int_example = 25

    @udf(
        (),
        {
            "int_col": Int,
            "int_col_32": Int32,
            "int_col_64": Int64,
            "float_col": Float,
            "float_col_32": Float32,
            "float_col_64": Float64,
            "array_col": Array(Float),
            "array_col_nested": Array(Array(Float)),
            "array_col_32": Array(Float32),
            "array_col_64": Array(Float64),
            "string_col": String,
            "bool_col": Boolean,
            "json_col": JSON,
            "binary_col": Binary,
        },
    )
    def test_types():
        return (
            5,
            5,
            5,
            0.5,
            0.5,
            0.5,
            [0.5],
            [[0.5], [0.5]],
            [0.5],
            [0.5],
            "s",
            True,
            dumps({"a": 1}),
            int_example.to_bytes(2, "big"),
        )

    (
        DatasetQuery(name=dogs_dataset.name, catalog=catalog)
        .add_signals(test_types)
        .save("dogs_custom_columns")
    )

    for r in catalog.get_dataset("dogs_custom_columns").get_version(1).preview:
        assert json.loads(r.get("custom")) == {
            "int_col": 5,
            "int_col_32": 5,
            "int_col_64": 5,
            "float_col": 0.5,
            "float_col_32": 0.5,
            "float_col_64": 0.5,
            "array_col": [0.5],
            "array_col_nested": [[0.5], [0.5]],
            "array_col_32": [0.5],
            "array_col_64": [0.5],
            "string_col": "s",
            "bool_col": True,
            "json_col": '{"a": 1}',
            "binary_col": [0, 25],
        }


def test_dataset_preview_last_modified(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog

    DatasetQuery(name=dogs_dataset.name, catalog=catalog).save("dogs_custom_columns")

    for r in catalog.get_dataset("dogs_custom_columns").get_version(1).preview:
        assert isinstance(r.get("last_modified"), str)


def test_merge_datasets(cloud_test_catalog, dogs_dataset, cats_dataset):
    catalog = cloud_test_catalog.catalog
    catalog.merge_datasets(
        cats_dataset,
        dogs_dataset,
        1,
        dst_version=2,
    )

    dogs_dataset = catalog.get_dataset(dogs_dataset.name)
    assert dogs_dataset.versions_values == [1, 2]

    assert {r.name for r in catalog.ls_dataset_rows(dogs_dataset.name, 2)} == {
        "dog1",
        "dog2",
        "dog3",
        "dog4",
        "cat1",
        "cat2",
    }

    cats_dep = catalog.get_dataset_dependencies(cats_dataset.name, 1)
    dogs_old_dep = catalog.get_dataset_dependencies(dogs_dataset.name, 1)
    dogs_new_dep = catalog.get_dataset_dependencies(dogs_dataset.name, 2)

    assert set(dogs_new_dep) == set(cats_dep + dogs_old_dep)


def test_merge_datasets_invalid_version(cloud_test_catalog, dogs_dataset, cats_dataset):
    catalog = cloud_test_catalog.catalog
    with pytest.raises(DatasetInvalidVersionError) as exc_info:
        catalog.merge_datasets(
            cats_dataset,
            dogs_dataset,
            1,
            dst_version=1,
        )
    assert str(exc_info.value) == (
        "Version 1 must be higher than the current latest one"
    )


def test_merge_datasets_existing_pending_version(
    cloud_test_catalog, dogs_dataset, cats_dataset
):
    catalog = cloud_test_catalog.catalog

    # adding custom column
    @udf(
        (),
        {"similarity": Float32},
    )
    def test_types():
        return (0.5,)

    (
        DatasetQuery(name=cats_dataset.name, catalog=catalog)
        .add_signals(test_types)
        .save("cats_custom_column")
    )

    cats_custom_column = catalog.get_dataset("cats_custom_column")

    dogs_dataset = catalog.create_new_dataset_version(
        dogs_dataset,
        2,
        create_rows_table=False,
    )

    catalog.merge_datasets(
        cats_custom_column,
        dogs_dataset,
        1,
        dst_version=2,
    )

    dogs_dataset = catalog.get_dataset(dogs_dataset.name)
    assert dogs_dataset.versions_values == [1, 2]

    assert (
        dogs_dataset.get_version(2).custom_column_types
        == cats_custom_column.get_version(1).custom_column_types
    )

    assert "similarity" in dogs_dataset.get_version(2).custom_column_types

    assert {r.name for r in catalog.ls_dataset_rows(dogs_dataset.name, 2)} == {
        "cat1",
        "cat2",
    }

    cats_dep = catalog.get_dataset_dependencies(cats_custom_column.name, 1)
    dogs_dep = catalog.get_dataset_dependencies(dogs_dataset.name, 2)

    assert set(dogs_dep) == set(cats_dep)


def test_merge_datasets_size(cloud_test_catalog, dogs_dataset, cats_dataset):
    catalog = cloud_test_catalog.catalog
    catalog.merge_datasets(
        cats_dataset,
        dogs_dataset,
        1,
        dst_version=2,
    )

    dogs_dataset = catalog.get_dataset(dogs_dataset.name)
    assert dogs_dataset.versions_values == [1, 2]

    dogs_version_1 = dogs_dataset.get_version(1)
    dogs_version_2 = dogs_dataset.get_version(2)
    cats_version_1 = cats_dataset.get_version(1)

    assert dogs_version_2.size
    assert dogs_version_2.size == dogs_version_1.size + cats_version_1.size

    assert dogs_version_2.num_objects
    assert (
        dogs_version_2.num_objects
        == dogs_version_1.num_objects + cats_version_1.num_objects
    )


@pytest.mark.parametrize("tree", [{str(i): str(i) for i in range(50)}], indirect=True)
def test_row_random(cloud_test_catalog):
    # Note: this is technically a probabilistic test, but the probability
    # of accidental failure is < 1e-10
    ctc = cloud_test_catalog
    catalog = ctc.catalog
    catalog.index([ctc.src_uri])
    catalog.create_dataset_from_sources("test", [ctc.src_uri])
    random_values = [row.random for row in catalog.ls_dataset_rows("test", 1)]

    # Random values are unique
    assert len(set(random_values)) == len(random_values)

    RAND_MAX = 2**63  # noqa: N806
    # Values are drawn uniformly from range(2**63)
    assert 0 <= min(random_values) < 0.4 * RAND_MAX
    assert 0.6 * RAND_MAX < max(random_values) < RAND_MAX

    # Creating a new dataset preserves random values
    catalog.create_dataset_from_sources("test2", [ctc.src_uri])
    random_values2 = {row.random for row in catalog.ls_dataset_rows("test2", 1)}
    assert random_values2 == set(random_values)


def test_dataset_stats_registered_ds(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog
    stats = catalog.dataset_stats(dogs_dataset.name, 1)
    assert stats.num_objects == 4
    assert stats.size == 15
    rows_count = catalog.warehouse.dataset_rows_count(dogs_dataset, 1)
    assert rows_count == 4


@pytest.mark.parametrize("indirect", [True, False])
def test_dataset_dependencies_registered(
    listed_bucket, cloud_test_catalog, dogs_dataset, indirect
):
    catalog = cloud_test_catalog.catalog
    storage = catalog.get_storage(cloud_test_catalog.storage_uri)

    assert [
        dataset_dependency_asdict(d)
        for d in catalog.get_dataset_dependencies(
            dogs_dataset.name, 1, indirect=indirect
        )
    ] == [
        {
            "id": ANY,
            "type": DatasetDependencyType.STORAGE,
            "name": storage.uri,
            "version": storage.timestamp_str,
            "created_at": isoparse(storage.timestamp_str),
            "dependencies": [],
        }
    ]



import posixpath as pp
import re

import pytest
from sqlalchemy import literal, select
from sqlalchemy.sql import func as f

from dvcx.sql.functions import path as sql_path

PATHS = ["", "/", "name", "/name", "name/", "some/long/path"]
EXT_PATHS = [
    "",
    "abc.txt",
    "abc...txt",
    "abc",
    "abc/",
    "some/path/abc.tar.gz",
    "some/pa.th/abc",
]


def split_parent(path):
    parent, name = f"/{path}".rsplit("/", 1)
    return parent[1:], name


def file_stem(path):
    return pp.splitext(path)[0].rstrip(".")


def file_ext(path):
    return pp.splitext(path)[1].lstrip(".")


@pytest.mark.parametrize("func_base", [f.path, sql_path])
@pytest.mark.parametrize("func_name", ["parent", "name"])
def test_default_not_implement(func_base, func_name):
    """
    Importing dvcx.sql.functions.path should register a custom compiler
    which raises an exception for these functions with the default
    SQLAlchemy dialect.
    """
    fn = getattr(func_base, func_name)
    expr = fn(literal("file:///some/file/path"))
    with pytest.raises(NotImplementedError, match=re.escape(f"path.{func_name}")):
        expr.compile()


@pytest.mark.parametrize("path", PATHS)
def test_parent(warehouse, path):
    query = select(f.path.parent(literal(path)))
    result = tuple(warehouse.db.execute(query))
    assert result == ((split_parent(path)[0],),)


@pytest.mark.parametrize("path", PATHS)
def test_name(warehouse, path):
    query = select(f.path.name(literal(path)))
    result = tuple(warehouse.db.execute(query))
    assert result == ((split_parent(path)[1],),)


@pytest.mark.parametrize("path", EXT_PATHS)
def test_file_stem(warehouse, path):
    query = select(sql_path.file_stem(literal(path)))
    result = tuple(warehouse.db.execute(query))
    assert result == ((file_stem(path),),)


@pytest.mark.parametrize("path", EXT_PATHS)
def test_file_ext(warehouse, path):
    query = select(sql_path.file_ext(literal(path)))
    result = tuple(warehouse.db.execute(query))
    assert result == ((file_ext(path),),)
